import toolviper.utils.logger as logger
import os
from pathlib import Path
import re
from typing import Any, Callable, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr

import astropy.units
from casacore import tables

from .table_query import open_query, open_table_ro
from xradio._utils.list_and_array import get_pad_value

CASACORE_TO_PD_TIME_CORRECTION = 3_506_716_800.0
SECS_IN_DAY = 86400
MJD_DIF_UNIX = 40587


def table_exists(path: str) -> bool:
    """
    Whether a casacore table exists on disk (in the casacore.tables.tableexists sense)
    """
    return tables.tableexists(path)


def table_has_column(path: str, column_name: str) -> bool:
    """
    Whether a column is present in a casacore table
    """
    with open_table_ro(path) as tb_tool:
        if column_name in tb_tool.colnames():
            return True
        else:
            return False


def convert_casacore_time(
    rawtimes: np.ndarray, convert_to_datetime: bool = True
) -> np.ndarray:
    """
    Read time columns to datetime format
    pandas datetimes are referenced against a 0 of 1970-01-01
    CASA's modified julian day reference time is (of course) 1858-11-17

    This requires a correction of 3506716800 seconds which is hardcoded to save time

    Parameters
    ----------
    rawtimes : np.ndarray
        times in casacore ref
    convert_to_datetime : bool (Default value = True)
        whether to produce pandas style datetime

    Returns
    -------
    np.ndarray
        times converted to pandas reference
    """
    times_reref = np.array(rawtimes) - CASACORE_TO_PD_TIME_CORRECTION
    if convert_to_datetime:
        return pd.to_datetime(times_reref, unit="s").values
    else:
        return times_reref
    # dt = pd.to_datetime(np.atleast_1d(rawtimes) - correction, unit='s').values
    # if len(np.array(rawtimes).shape) == 0: dt = dt[0]
    # return dt


def convert_mjd_time(rawtimes: np.ndarray) -> np.ndarray:
    """
    Different time conversion needed for the MJD col of EPHEM{i}_*.tab
    files (only, as far as I've seen)

    Parameters
    ----------
    rawtimes : np.ndarray
        MJD times for example from the MJD col of ephemerides tables

    Returns
    -------
    np.ndarray
        times converted to pandas reference and datetime type
    """
    times_reref = pd.to_datetime(
        (rawtimes - MJD_DIF_UNIX) * SECS_IN_DAY, unit="s"
    ).values

    return times_reref


def convert_casacore_time_to_mjd(rawtimes: np.ndarray) -> np.ndarray:
    """
    From CASA/casacore time (as used in the TIME column of the main table) to MJD
    (as used in the EPHEMi*.tab ephemeris tables). As the epochs are the same, this
    is just a conversion of units.

    Parameters
    ----------
    rawtimes : np.ndarray
        times from a TIME column (seconds, casacore time epoch)

    Returns
    -------
    np.ndarray
        times converted to (ephemeris) MJD (days since casacore time epoch (1858-11-17))
    """
    return rawtimes / SECS_IN_DAY


def make_taql_where_between_min_max(
    min_max: Tuple[np.float64, np.float64],
    path: str,
    table_name: str,
    colname="TIME",
) -> Union[str, None]:
    """
    From a numerical min/max range, produce a TaQL string to select between
    those min/max values (example: times) in a table.
    The table can be for example a POINTING subtable or an EPHEM* ephemeris
    table.
    This is meant to be used on MSv2 table columns that will be loaded as a
    coordinate in MSv4s and their sub-xdss (example: POINTING/TIME ephemeris/MJD).

    This can be used for example to produce a TaQL string to constraing loading of:
    - POINTING rows (based on the min/max from the time coordinate of the main MSv4)
    - ephemeris rows, from EPHEM* tables ((based on the MJD column and the min/max
      from the main MSv4 time coordinate).

    Parameters
    ----------
    min_max : Tuple[np.float64, np.float64]
        min / max values of time or other column used as coordinate
        (assumptions: float values, sortable, typically: time coord from MSv4)
    path :
        Path to input MS or location of the table
    table_name :
        Name of the table where to load a column (example: 'POINTING')
    colname :
        Name of the column to search for min/max values (examples: 'TIME', 'MJD')

    Returns
    -------
    taql_where : str
        TaQL (sub)string with the min/max time 'WHERE' constraint
    """

    min_max_range = find_projected_min_max_table(min_max, path, table_name, colname)
    if min_max_range is None:
        taql = None
    else:
        (min_val, max_val) = min_max_range
        taql = f"where {colname} >= {min_val} AND {colname} <= {max_val}"

    return taql


def find_projected_min_max_table(
    min_max: Tuple[np.float64, np.float64], path: str, table_name: str, colname: str
) -> Union[Tuple[np.float64, np.float64], None]:
    """
    We have: min/max values that define a range (for example of time)
    We want: to project that min/max range on a sortable column (for example a
    range of times onto a TIME column), and find min and max values
    derived from that table column such that the range between those min and max
    values includes at least the input min/max range.

    The returned min/max can then be used in a data selection TaQL query to
    select at least the values within the input range (possibly extended if
    the input range overlaps only partially or not at all with the column
    values). A tolerance is added to the min/max to prevent numerical issues in
    comparisons and conversios between numerical types and strings.

    When the range given as input is wider than the range of values found in
    the column, use the input range, as it is sufficient and more inclusive.

    When the range given as input is narrow (projected on the target table/column)
    and falls between two points of the column values, or overlaps with only one
    point, the min/max are extended to include at least the two column values that
    define a range within which the input range is included.
    Example scenario: an ephemeris table is sampled at a coarse interval
    (20 min) and we want to find a min/max range projected from the time min/max
    of a main MSv4 time coordinate sampled at ~1s for a field-scan/intent
    that spans ~2 min. Those ~2 min will typically fall between ephemeris samples.

    Parameters
    ----------
    min_max : Tuple[np.float64, np.float64]
        min / max values of time or other column used as coordinate
        (assumptions: float values, sortable)
    path :
        Path to input MS or location of the table
    table_name :
        Name of the table where to load a column (example: 'POINTING')
    colname :
        Name of the column to search for min/max values (example: 'TIME')

    Returns
    -------
    output_min_max : Union[Tuple[np.float64, np.float64], None]
        min/max values derived from the input min/max and the column values
    """
    with open_table_ro(os.path.join(path, table_name)) as tb_tool:
        if tb_tool.nrows() == 0:
            return None
        col = tb_tool.getcol(colname)

    out_min_max = find_projected_min_max_array(min_max, col)
    return out_min_max


def find_projected_min_max_array(
    min_max: Tuple[np.float64, np.float64], array: np.array
) -> Tuple[np.float64, np.float64]:
    """Does the min/max checks and search for find_projected_min_max_table()"""

    sorted_array = np.sort(array)
    (range_min, range_max) = min_max
    if len(sorted_array) < 2:
        tol = np.finfo(sorted_array.dtype).eps * 4
    else:
        tol = np.diff(sorted_array[np.nonzero(sorted_array)]).min() / 4

    if range_max > sorted_array[-1]:
        projected_max = range_max + tol
    else:
        max_idx = sorted_array.size - 1
        max_array_idx = min(
            max_idx, np.searchsorted(sorted_array, range_max, side="right")
        )
        projected_max = sorted_array[max_array_idx] + tol

    if range_min < sorted_array[0]:
        projected_min = range_min - tol
    else:
        min_array_idx = max(
            0, np.searchsorted(sorted_array, range_min, side="left") - 1
        )
        # ensure 'sorted_array[min_array_idx] < range_min' when values ==
        if sorted_array[min_array_idx] == range_min:
            min_array_idx = max(0, min_array_idx - 1)
        projected_min = sorted_array[min_array_idx] - tol

    return (projected_min, projected_max)


def extract_table_attributes(infile: str) -> Dict[str, Dict]:
    """
    Return a dictionary of table attributes created from MS keywords and column descriptions

    Parameters
    ----------
    infile : str
        table file path

    Returns
    -------
    Dict[str, Dict]
        table attributes as a dictionary
    """
    with open_table_ro(infile) as tb_tool:
        kwd = tb_tool.getkeywords()
        attrs = dict([(kk, kwd[kk]) for kk in kwd if kk not in os.listdir(infile)])
        cols = tb_tool.colnames()
        column_descriptions = {}
        for col in cols:
            column_descriptions[col] = tb_tool.getcoldesc(col)
        attrs["column_descriptions"] = column_descriptions
        attrs["info"] = tb_tool.info()

    return attrs


def add_units_measures(
    mvars: Dict[str, xr.DataArray], cc_attrs: Dict[str, Any]
) -> Dict[str, xr.DataArray]:
    """
    Add attributes with units and measure metainfo to the variables passed in the input dictionary

    Parameters
    ----------
    mvars : Dict[str, xr.DataArray]
        data variables where to populate units
    cc_attrs : Dict[str, Any]
        dictionary with casacore table attributes (from extract_table_attributes)

    Returns
    -------
    Dict[str, xr.DataArray]
        variables with units added in their attributes
    """
    col_descrs = cc_attrs["column_descriptions"]
    # TODO: Should probably loop the other way around, over mvars
    for col in col_descrs:
        if col == "TIME":
            var_name = "time"
        else:
            var_name = col
        if var_name in mvars and "keywords" in col_descrs[col]:
            if "QuantumUnits" in col_descrs[col]["keywords"]:
                cc_units = col_descrs[col]["keywords"]["QuantumUnits"]

                if isinstance(
                    cc_units, str
                ):  # Little fix for Meerkat data where the units are a string.
                    cc_units = [cc_units]

                if not isinstance(cc_units, list) or not cc_units:
                    logger.warning(
                        f"Invalid units found for column/variable {col}: {cc_units}"
                    )
                mvars[var_name].attrs["units"] = cc_units[0]
                try:
                    astropy.units.Unit(cc_units[0])
                except Exception as exc:
                    logger.warning(
                        f"Unsupported units found for column/variable {col}: "
                        f"{cc_units}. Cannot create an astropy.units.Units object from it: {exc}"
                    )

            if "MEASINFO" in col_descrs[col]["keywords"]:
                cc_meas = col_descrs[col]["keywords"]["MEASINFO"]
                mvars[var_name].attrs["measure"] = {"type": cc_meas["type"]}
                # VarRefCol used for several cols of:
                # - SPECTRAL_WINDOW (MEAS_FREG_REF, in MSv2 std)
                # - FIELD  (PhseDir_Ref, DelayDir_Ref, RefDir_Ref, not in MSv2 std)
                # - POINTING - to be split
                if "VarRefCol" not in cc_meas:
                    mvars[var_name].attrs["measure"]["ref_frame"] = cc_meas["Ref"]
                else:
                    mvars[var_name].attrs["measure"]["ref_frame_data_var"] = cc_meas[
                        "VarRefCol"
                    ]
                    if "TabRefTypes" in cc_meas:
                        mvars[var_name].attrs["measure"].update(
                            {
                                "ref_frame_types": list(cc_meas["TabRefTypes"]),
                                "ref_frame_codes": list(cc_meas["TabRefCodes"]),
                            }
                        )

    return mvars


def make_freq_attrs(spw_xds: xr.Dataset, spw_id: int) -> Dict[str, Any]:
    """
    Grab the units/measure metainfo for the xds.freq dimension of a
    parttion from the SPECTRAL_WINDOW subtable CTDS attributes.

    Has to read xds_spw.meas_freq_ref and use it as index in the CTDS
    'VarRefCol' attrs of CHAN_FREQ and REF_FREQUENCY to give a
    reference frame to xds_spw.ref_frequency and xds_spw.chan_freq
    (then the ref frame from the second will be pulled to
    xds.freq.attrs)

    Parameters
    ----------
    spw_xds : xr.Dataset
        (metainfo) SPECTRAL_WINDOW xds
    spw_id : int
        SPW id of a partition

    Returns
    -------
    Dict[str, Any]
        attributes (units/measure) for the freq dim of a partition
    """
    fallback_TabRefTypes = [
        "REST",
        "LSRK",
        "LSRD",
        "BARY",
        "GEO",
        "TOPO",
        "GALACTO",
        "LGROUP",
        "CMB",
    ]

    ctds_cols = spw_xds.attrs["other"]["msv2"]["ctds_attrs"]["column_descriptions"]
    cfreq = ctds_cols["CHAN_FREQ"]

    cf_attrs = spw_xds.data_vars["CHAN_FREQ"].attrs
    if "MEASINFO" in cfreq["keywords"] and "VarRefCol" in cfreq["keywords"]["MEASINFO"]:
        fattrs = cfreq["keywords"]["MEASINFO"]
        var_ref_col = fattrs["VarRefCol"]
        # This should point to the SPW/MEAS_FREQ_REF col
        meas_freq_ref_idx = spw_xds.data_vars[var_ref_col].values[spw_id]

        if "TabRefCodes" not in fattrs or "TabRefTypes" not in fattrs:
            # Datasets like vla/ic2233_1.ms say "VarRefCol" but "TabRefTypes" is missing
            ref_frame = fallback_TabRefTypes[meas_freq_ref_idx]
        else:
            ref_type_code = fattrs["TabRefCodes"][meas_freq_ref_idx]
            ref_frame = fattrs["TabRefTypes"][ref_type_code]

        cf_attrs["measure"] = {
            "type": fattrs["type"],
            "ref_frame": ref_frame,
        }

        # Also set the 'VarRefCol' for CHAN_FREQ and REF_FREQUENCEY
        spw_xds.data_vars["CHAN_FREQ"].attrs.update(cf_attrs)
        spw_xds.data_vars["REF_FREQUENCY"].attrs.update(cf_attrs)

    return cf_attrs


def redimension_ms_subtable(xds: xr.Dataset, subt_name: str) -> xr.Dataset:
    """
    Expand a MeasurementSet subtable xds from single dimension (row)
    to multiple dimensions (such as (source_id, time, spectral_window)

    WIP: only works for source, experimenting

    Parameters
    ----------
    xds : xr.Dataset
        dataset to change the dimensions
    subt_name : str
        subtable name (SOURCE, etc.)

    Returns
    -------
    xr.Dataset
        transformed xds with data dimensions representing the MS subtable key
        (one dimension for every columns)
    """
    subt_key_cols = {
        "DOPPLER": ["DOPPLER_ID", "SOURCE_ID"],
        "FREQ_OFFSET": [
            "ANTENNA1",
            "ANTENNA2",
            "FEED_ID",
            "SPECTRAL_WINDOW_ID",
            "TIME",
        ],
        "POINTING": ["TIME", "ANTENNA_ID"],
        "SOURCE": ["SOURCE_ID", "TIME", "SPECTRAL_WINDOW_ID"],
        "SYSCAL": ["ANTENNA_ID", "FEED_ID", "SPECTRAL_WINDOW_ID", "TIME"],
        "WEATHER": ["ANTENNA_ID", "TIME"],
        "PHASE_CAL": ["ANTENNA_ID", "TIME", "SPECTRAL_WINDOW_ID"],
        "GAIN_CURVE": ["ANTENNA_ID", "TIME", "SPECTRAL_WINDOW_ID"],
        "FEED": ["ANTENNA_ID", "SPECTRAL_WINDOW_ID"],
        # added tables (MSv3 but not preent in MSv2). Build it from "EPHEMi_... tables
        # Not clear what to do about 'time' var/dim:  , "time"],
        "EPHEMERIDES": ["ephemeris_row_id", "ephemeris_id"],
    }
    key_dims = subt_key_cols[subt_name]

    rxds = xds.copy()
    try:
        # drop_duplicates() needed (https://github.com/casangi/xradio/issues/185). Examples:
        # - Some early ALMA datasets have bogus WEATHER tables with many/most rows with
        #   (ANTENNA_ID=0, TIME=0) and no other columns to figure out the right IDs, such
        #   as "NS_WX_STATION_ID" or similar. (example: X425.pm04.scan4.ms)
        # - Some GBT MSs have duplicated (ANTENNA_ID=0, TIME=xxx). (example: analytic_variable.ms)
        rxds = (
            rxds.set_index(row=key_dims)
            .drop_duplicates("row")
            .unstack("row")
            .transpose(*key_dims, ...)
        )
        # unstack changes type to float when it needs to introduce NaNs, so
        # we need to reset to the original type.
        for var in rxds.data_vars:
            if rxds[var].dtype != xds[var].dtype:
                # beware of gaps/empty==nan values when redimensioning
                with np.errstate(invalid="ignore"):
                    rxds[var] = rxds[var].astype(xds[var].dtype)
    except Exception as exc:
        logger.warning(
            f"Cannot expand rows in table {subt_name} to {key_dims}, possibly duplicate values in those coordinates. "
            f"Exception: {exc}"
        )
        rxds = xds.copy()

    return rxds


def is_ephem_subtable(tname: str) -> bool:
    return "EPHEM" in tname and Path(tname).name.startswith("EPHEM")


def add_ephemeris_vars(tname: str, xds: xr.Dataset) -> xr.Dataset:
    fname = Path(tname).name
    pattern = r"EPHEM(\d+)_"
    match = re.match(pattern, fname)
    if match:
        ephem_id = match.group(1)
    else:
        ephem_id = 0

    xds["ephemeris_id"] = np.uint32(ephem_id) * xr.ones_like(
        xds["MJD"], dtype=np.uint32
    )
    xds = xds.rename({"MJD": "time"})
    xds["ephemeris_row_id"] = (
        xr.zeros_like(xds["time"], dtype=np.uint32) + xds["row"].values
    )

    return xds


def is_nested_ms(attrs: Dict) -> bool:
    ctds_attrs = attrs["other"]["msv2"]["ctds_attrs"]
    return (
        "MS_VERSION" in ctds_attrs
        and "column_descriptions" in ctds_attrs
        and all(
            col in ctds_attrs["column_descriptions"]
            for col in (
                "UVW",
                "ANTENNA1",
                "ANTENNA2",
                "FEED1",
                "FEED2",
                "OBSERVATION_ID",
            )
        )
    )


def load_generic_table(
    inpath: str,
    tname: str,
    timecols: Union[List[str], None] = None,
    ignore: Union[List[str], None] = None,
    rename_ids: Dict[str, str] = None,
    taql_where: str = None,
) -> xr.Dataset:
    """
    load generic casacore (sub)table into memory resident xds (xarray wrapped
    numpy arrays). This reads through the table columns and loads the data.

    TODO: change read_ name to load_ name (and most if not all this module)

    Parameters
    ----------
    inpath : str
        path to the MS or directory containing the table
    tname : str
        (sub)table name, for example 'SOURCE' for myms.ms/SOURCE
    timecols : Union[List[str], None] (Default value = None)
        column names to convert to numpy datetime format.
        leaves times as their original casacore format.
    ignore : Union[List[str], None] (Default value = None)
        list of column names to ignore and not try to read.
    rename_ids : Dict[str, str] (Default value = None)
        dict with dimension renaming mapping
    taql_where : str (Default value = None)
         TaQL string to optionally constain the rows/columns to read
         (Default value = None)

    Returns
    -------
    xr.Dataset
        table loaded as XArray dataset
    """
    if timecols is None:
        timecols = []
    if ignore is None:
        ignore = []

    infile = Path(inpath, tname)
    infile = str(infile.expanduser())
    if not os.path.isdir(infile):
        raise ValueError(
            f"invalid input filename to load_generic_table: {infile} table {tname}"
        )

    cc_attrs = extract_table_attributes(infile)
    attrs: Dict[str, Any] = {"other": {"msv2": {"ctds_attrs": cc_attrs}}}
    if is_nested_ms(attrs):
        logger.warning(
            f"Skipping subtable that looks like a MeasurementSet main table: {inpath} {tname}"
        )
        return xr.Dataset()

    with open_table_ro(infile) as gtable:
        if gtable.nrows() == 0:
            logger.debug(f"table is empty: {inpath} {tname}")
            return xr.Dataset(attrs=attrs)

        # if len(ignore) > 0: #This is needed because some SOURCE tables have a SOURCE_MODEL column that is corrupted and this causes the open_query to fail.
        #     select_columns = gtable.colnames()
        #     select_columns_str = str([item for item in select_columns if item not in ignore])[1:-1].replace("'", "") #Converts an array to a comma sepearted string. For example ['a', 'b', 'c'] to 'a, b, c'.
        #     taql_gtable = f"select " + select_columns_str + f" from $gtable {taql_where}"
        # else:
        #     taql_gtable = f"select * from $gtable {taql_where}"

        # relatively often broken columns that we do not need
        exclude_pattern = ", !~p/SOURCE_MODEL/"
        taql_gtable = f"select *{exclude_pattern} from $gtable {taql_where or ''}"

        with open_query(gtable, taql_gtable) as tb_tool:
            if tb_tool.nrows() == 0:
                logger.debug(
                    f"table query is empty: {inpath} {tname}, with where {taql_where}"
                )
                return xr.Dataset(attrs=attrs)

            colnames = tb_tool.colnames()
            mcoords, mvars = load_cols_into_coords_data_vars(
                infile, tb_tool, timecols, ignore
            )

    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    xds = xr.Dataset(mvars, coords=mcoords)

    dim_prefix = "dim"
    dims = ["row"] + [f"{dim_prefix}_{i}" for i in range(1, 20)]
    xds = xds.rename(dict([(dv, dims[di]) for di, dv in enumerate(xds.sizes)]))
    if rename_ids:
        rename_ids = {k: v for k, v in rename_ids.items() if k in xds.sizes}
    xds = xds.rename_dims(rename_ids)

    attrs["other"]["msv2"]["bad_cols"] = list(
        np.setdiff1d(
            [dv for dv in colnames],
            [dv for dv in list(xds.data_vars) + list(xds.coords)],
        )
    )

    if tname in [
        "DOPPLER",
        "FREQ_OFFSET",
        "POINTING",
        "SOURCE",
        "SYSCAL",
        "WEATHER",
        "PHASE_CAL",
        "GAIN_CURVE",
        "FEED",
    ]:
        xds = redimension_ms_subtable(xds, tname)

    if is_ephem_subtable(tname):
        xds = add_ephemeris_vars(tname, xds)
        xds = redimension_ms_subtable(xds, "EPHEMERIDES")

    xds = xds.assign_attrs(attrs)

    return xds


def load_cols_into_coords_data_vars(
    inpath: str,
    tb_tool: tables.table,
    timecols: Union[List[str], None] = None,
    ignore: Union[List[str], None] = None,
) -> Tuple[Dict[str, xr.Dataset], Dict[str, xr.Dataset]]:
    """
    Produce a set of coordinate xarrays and a set of data variables xarrays
    from the columns of a table.

    Parameters
    ----------
    inpath : str
        input path
    tb_tool: tables.table
        tool being used to load data
    timecols: Union[List[str], None] (Default value = None)
        list of columns to be considered as TIME-related
    ignore: Union[List[str], None] (Default value = None)
        columns to ignore

    Returns
    -------
    Tuple[Dict[str, xr.Dataset], Dict[str, xr.Dataset]]
        coordinates dictionary + variables dictionary
    """
    columns_loader = find_best_col_loader(inpath, tb_tool.nrows())

    mcoords, mvars = columns_loader(inpath, tb_tool, timecols, ignore)

    return mcoords, mvars


def find_best_col_loader(inpath: str, nrows: int) -> Callable:
    """
    Simple heuristic: for any tables other than POINTING, use the generic_load_cols
    function that is able to deal with variable size columns. For POINTING (and if it has
    more rows than an arbitrary "small" threshold) use a more efficient load function that
    loads the data by column (but is not able to deal with any generic table).
    For now, all other subtables are loaded using the generic column loader.

    Background: the POINTING subtable can have a very large number of rows. For example in
    ALMA it is sampled at ~50ms intervals which typically produces of the order of
    [10^5, 10^7] rows. This becomes a serious performance bottleneck when loading the
    table using row() (and one dict allocated per row).
    This function chooses an alternative "by-column" load function to load in the columns
    when the table is POINTING. See xradio issue #128 for now this distinction is made
    solely for performance reasons.

    Parameters
    ----------
    inpath : str
        path name of the MS table
    nrows : int
        number of rows found in the table

    Returns
    -------
    Callable
        function best suited to load the data from the columns of this table
    """
    # do not give up generic by-row() loading if nrows is (arbitrary) small
    ARBITRARY_MIN_ROWS = 1000

    if inpath.endswith("POINTING") and nrows >= ARBITRARY_MIN_ROWS:
        columns_loader = load_fixed_size_cols
    else:
        columns_loader = load_generic_cols

    return columns_loader


def load_generic_cols(
    inpath: str,
    tb_tool: tables.table,
    timecols: Union[List[str], None],
    ignore: Union[List[str], None],
) -> Tuple[Dict[str, xr.Dataset], Dict[str, xr.Dataset]]:
    """
    Loads data for each MS column (loading the data in memory) into Xarray datasets

    This function is generic in that it can load variable size array columns. See also
    load_fixed_size_cols() as a simpler and much better performing alternative
    for tables that are large and expected/guaranteed to not have columns with variable
    size cells.

    Parameters
    ----------
    inpath : str
        path name of the MS table
    tb_tool : tables.table
        table to load the columns
    timecols : Union[List[str], None]
        columns names to convert to datetime format
    ignore : Union[List[str], None]
        list of column names to skip and not try to load.

    Returns
    -------
    Tuple[Dict[str, xr.Dataset], Dict[str, xr.Dataset]]
        dict of coordinates and dict of data vars.
    """

    col_types = find_loadable_cols(tb_tool, ignore)

    trows = tb_tool.row(ignore, exclude=True)[:]

    # Produce coords and data vars from MS columns
    mcoords, mvars = {}, {}
    for col in col_types.keys():
        try:
            # TODO
            # benchmark np.stack() performance
            data = np.stack(
                [row[col] for row in trows]
            )  # .astype(col_cells[col].dtype)
            if isinstance(trows[0][col], dict):
                # TODO
                # benchmark np.stack() performance
                data = np.stack(
                    [
                        (
                            row[col]["array"].reshape(row[col]["shape"])
                            if len(row[col]["array"]) > 0
                            else np.array([""])
                        )
                        for row in trows
                    ]
                )
        except Exception:
            # sometimes the cols are variable, so we need to standardize to the largest sizes

            if len(set([isinstance(row[col], dict) for row in trows])) > 1:
                continue  # can't deal with this case

            data = handle_variable_col_issues(inpath, col, col_types[col], trows)

        if len(data) == 0:
            continue

        array_type, array_data = raw_col_data_to_coords_vars(
            inpath, tb_tool, col, data, timecols
        )
        if array_type == "coord":
            mcoords[col] = array_data
        elif array_type == "data_var":
            mvars[col] = array_data

    return mcoords, mvars


def load_fixed_size_cols(
    inpath: str,
    tb_tool: tables.table,
    timecols: Union[List[str], None],
    ignore: Union[List[str], None],
) -> Tuple[Dict[str, xr.Dataset], Dict[str, xr.Dataset]]:
    """
    Loads columns into memory via the table tool getcol() function, as opposed to
    load_generic_cols() which loads on a per-row basis via row().
    This function is 2+ orders of magnitude faster for large tables (pointing tables with
    the order of >=10^5 rows)
    Prefer this function for performance reasons when all rows can be assumed to be fixed
    size (even if they are of array type).
    This is performance-critical for the POINTING subtable.

    Parameters
    ----------
    inpath : str
        path name of the MS
    tb_tool : tables.table
        table to red the columns
    timecols : Union[List[str], None]
        columns names to convert to datetime format
    ignore : Union[List[str], None]
        list of column names to skip and not try to load.

    Returns
    -------
    Tuple[Dict[str, xr.Dataset], Dict[str, xr.Dataset]]
        dict of coordinates and dict of data vars, ready to construct an xr.Dataset
    """

    loadable_cols = find_loadable_cols(tb_tool, ignore)

    # Produce coords and data vars from MS columns
    mcoords, mvars = {}, {}
    for col in loadable_cols.keys():
        try:
            data = tb_tool.getcol(col)
            if isinstance(data, dict):
                data = data["array"].reshape(data["shape"])
        except Exception as exc:
            logger.warning(
                f"{inpath}: failed to load data with getcol for column {col}: {exc}"
            )
            data = []

        if len(data) == 0:
            continue

        array_type, array_data = raw_col_data_to_coords_vars(
            inpath, tb_tool, col, data, timecols
        )
        if array_type == "coord":
            mcoords[col] = array_data
        elif array_type == "data_var":
            mvars[col] = array_data

    return mcoords, mvars


def find_loadable_cols(
    tb_tool: tables.table, ignore: Union[List[str], None]
) -> Dict[str, str]:
    """
    For a table, finds the columns that are loadable = not of record type,
    and not to be ignored
    In extreme cases of variable size columns, it can happen that all the
    cells are empty (iscelldefined() == false). This is still considered a
    loadable column, even though all values of the resulting data var will
    be empty.

    Parameters
    ----------
    tb_tool : tables.table
        table to red the columns
    ignore : Union[List[str], None]
        list of column names to skip and not try to load.

    Returns
    -------
    Dict
        dict of {column name: column type} for columns that can/should be loaded
    """

    colnames = tb_tool.colnames()
    table_desc = tb_tool.getdesc()
    loadable_cols = {
        col: table_desc[col]["valueType"]
        for col in colnames
        if (col not in ignore) and tb_tool.coldatatype(col) != "record"
    }
    return loadable_cols


def raw_col_data_to_coords_vars(
    inpath: str,
    tb_tool: tables.table,
    col: str,
    data: np.ndarray,
    timecols: Union[List[str], None],
) -> Tuple[str, xr.DataArray]:
    """
    From a raw np array of data (freshly loaded from a table column), prepares either a
    coord or a data_var ready to be added to an xr.Dataset

    Parameters
    ----------
    inpath: str
        input table path
    tb_tool: tables.table :
        table toold being used to load data
    col: str :
        column
    data: np.ndarray :
        column data
    timecols: Union[List[str], None]
        columns to be treated as TIME-related

    Returns
    -------
    Tuple[str, xr.DataArray]
        array type string (whether this column is a 'coord' or a 'data_var') + DataArray
        with column  data/coord values ready to be added to the table xds
    """

    # Almost sure that when TIME is present (in a standard MS subt) it
    # is part of the key. But what about non-std subtables, ASDM subts?
    subts_with_time_key = (
        "FLAG_CMD",
        "FREQ_OFFSET",
        "HISTORY",
        "POINTING",
        "SOURCE",
        "SYSCAL",
        "WEATHER",
        "PHASE_CAL",
        "GAIN_CURVE",
        "FEED",
    )
    dim_prefix = "dim"

    if col in timecols:
        if col == "MJD":
            data = convert_mjd_time(data).astype("float64") / 1e9
        else:
            try:
                data = convert_casacore_time(data)
            except pd.errors.OutOfBoundsDatetime as exc:
                if inpath.endswith("WEATHER"):
                    # intentionally not callling logging.exception
                    logger.warning(
                        f"Exception when converting WEATHER/TIME: {exc}. TIME data: {data}"
                    )
                else:
                    raise
    # should also probably add INTERVAL not only TIME
    if col.endswith("_ID") or (inpath.endswith(subts_with_time_key) and col == "TIME"):
        # weather table: importasdm produces very wrong "-1" ANTENNA_ID
        if (
            inpath.endswith("WEATHER")
            and col == "ANTENNA_ID"
            and "NS_WX_STATION_ID" in tb_tool.colnames()
        ):
            data = tb_tool.getcol("NS_WX_STATION_ID")

        array_type = "coord"
        array_data = xr.DataArray(
            data,
            dims=[
                f"{dim_prefix}_{di}_{ds}" for di, ds in enumerate(np.array(data).shape)
            ],
        )
    else:
        array_type = "data_var"
        array_data = xr.DataArray(
            data,
            dims=[
                f"{dim_prefix}_{di}_{ds}" for di, ds in enumerate(np.array(data).shape)
            ],
        )

    return array_type, array_data


def get_pad_value_in_tablerow_column(trows: tables.tablerow, col: str) -> object:
    """
    Gets the pad value for the type of a column (IMPORTANTLY) as froun in the
    the type specified in the row / column value dict returned by tablerow.
    This can differ from the type of the column as given in the casacore
    column descriptions. See https://github.com/casangi/xradio/issues/242.

    Parameters
    ----------
    trows : tables.tablerow
        list of rows from a table as loaded by tables.row()
    col: str
        get the pad value for this column

    Returns
    -------
    object
        pad value as produced by get_pad_value for the appropriate data type from
    tablerow
    """
    col_value = trows[0][col]
    if isinstance(col_value, np.ndarray):
        col_dtype = col_value.dtype
    elif isinstance(col_value, list):
        col_dtype = type(col_value[0])
    else:
        raise RuntimeError(
            "Found unexpected type (not np.array or list) in column value of "
            f"first row of column {col}: {col_value}"
        )

    return get_pad_value(col_dtype)


def handle_variable_col_issues(
    inpath: str, col: str, col_type: str, trows: tables.tablerow
) -> np.ndarray:
    """
    load variable-size array columns, padding with missing/fill/nans
    wherever needed. This happens for example often in the
    SPECTRAL_WINDOW table (CHAN_WIDTH, EFFECTIVE_BW, etc.).
    Also handle exceptions gracefully when trying to load the rows.

    Parameters
    ----------
    inpath : str
        path name of the MS
    col : str
        column being loaded
    col_type : str
        type of the column cell values (as numpy dtype string)
    trows : tables.tablerow
        rows from a table as loaded by tables.row()

    Returns
    -------
    np.ndarray
        array with column values (possibly padded if rows vary in size)
    """

    # Optional cols known to sometimes have inconsistent values
    known_misbehaving_cols = ["ASSOC_NATURE"]

    mshape = np.array(max([np.array(row[col]).shape for row in trows]))
    try:
        pad_val = None
        pad_val = get_pad_value_in_tablerow_column(trows, col)

        # TODO
        # benchmark np.stack() performance
        data = np.stack(
            [
                np.pad(
                    (
                        row[col]
                        if len(row[col]) > 0
                        else np.array(row[col]).reshape(np.arange(len(mshape)) * 0)
                    ),
                    [(0, ss) for ss in mshape - np.array(row[col]).shape],
                    "constant",
                    constant_values=pad_val,
                )
                for row in trows
            ]
        )
    except Exception as exc:
        msg = f"{inpath}: failed to load data for column {col}, with {pad_val=}: {exc}"
        if col in known_misbehaving_cols:
            logger.debug(msg)
        else:
            logger.warning(msg)
        data = np.empty(0)

    return data


def read_flat_col_chunk(infile, col, cshape, ridxs, cstart, pstart) -> np.ndarray:
    """
    Extract data chunk for each table col, this is fed to dask.delayed

    Parameters
    ----------
    infile :

    col :

    cshape :

    ridxs :

    cstart :

    pstart :

    Returns
    -------
    np.ndarray
    """

    with open_table_ro(infile) as tb_tool:
        rgrps = [
            (rr[0], rr[-1])
            for rr in np.split(ridxs, np.where(np.diff(ridxs) > 1)[0] + 1)
        ]
        # try:
        if (len(cshape) == 1) or (col == "UVW"):  # all the scalars and UVW
            data = np.concatenate(
                [tb_tool.getcol(col, rr[0], rr[1] - rr[0] + 1) for rr in rgrps], axis=0
            )
        elif len(cshape) == 2:  # WEIGHT, SIGMA
            data = np.concatenate(
                [
                    tb_tool.getcolslice(
                        col,
                        pstart,
                        pstart + cshape[1] - 1,
                        [],
                        rr[0],
                        rr[1] - rr[0] + 1,
                    )
                    for rr in rgrps
                ],
                axis=0,
            )
        elif len(cshape) == 3:  # DATA and FLAG
            data = np.concatenate(
                [
                    tb_tool.getcolslice(
                        col,
                        (cstart, pstart),
                        (cstart + cshape[1] - 1, pstart + cshape[2] - 1),
                        [],
                        rr[0],
                        rr[1] - rr[0] + 1,
                    )
                    for rr in rgrps
                ],
                axis=0,
            )
        # except:
        #    print('ERROR reading chunk: ', col, cshape, cstart, pstart)

    return data


def read_col_chunk(
    infile: str,
    ts_taql: str,
    col: str,
    cshape: Tuple[int],
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    didxs: np.ndarray,
    d1: Tuple[int, int],
    d2: Tuple[int, int],
) -> np.ndarray:
    """
    Function to perform delayed reads from table columns.

    Parameters
    ----------
    infile : str

    ts_taql : str

    col : str

    cshape : Tuple[int]

    tidxs : np.ndarray

    bidxs : np.ndarray

    didxs : np.ndarray

    d1: Tuple[int, int]

    d2: Tuple[int, int]

    Returns
    -------
    np.ndarray
    """
    # TODO: consider calling load_col_chunk() from inside the withs
    # for read_delayed_pointing_table and read_expanded_main_table
    with open_table_ro(infile) as mtable:
        with open_query(mtable, ts_taql) as query:
            if (len(cshape) == 2) or (col == "UVW"):  # all the scalars and UVW
                data = np.array(query.getcol(col, 0, -1))
            elif len(cshape) == 3:  # WEIGHT, SIGMA
                data = query.getcolslice(col, d1[0], d1[1], [], 0, -1)
            elif len(cshape) == 4:  # DATA and FLAG
                data = query.getcolslice(col, (d1[0], d2[0]), (d1[1], d2[1]), [], 0, -1)

    fill_value = get_pad_value(data.dtype)
    fulldata = np.full(cshape, fill_value, dtype=data.dtype)

    if len(didxs) > 0:
        fulldata[tidxs[didxs], bidxs[didxs]] = data[didxs]

    return fulldata


def read_col_conversion(
    tb_tool: tables.table,
    col: str,
    cshape: Tuple[int],
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    use_table_iter: bool,
) -> np.ndarray:
    """
    Function to perform delayed reads from table columns when converting
    (no need for didxs)

    Parameters
    ----------
    tb_tool : tables.table

    col : str

    cshape : Tuple[int]

    tidxs : np.ndarray

    bidxs : np.ndarray

    Returns
    -------
    np.ndarray
    """

    # Workaround for https://github.com/casacore/python-casacore/issues/130
    # WARNING: Assumes tb_tool is a single measurement set not an MMS.
    # WARNING: Assumes the num_frequencies * num_polarizations < 2**29. If false,
    # https://github.com/casacore/python-casacore/issues/130 isn't mitigated.

    # Use casacore to get the shape of a row for this column
    #################################################################################

    # Get the total number of rows in the base measurement set
    nrows_total = tb_tool.nrows()

    # getcolshapestring() only works on columns where a row element is an
    # array ie. fails for TIME
    # Assumes the RuntimeError is because the column is a scalar
    try:
        shape_string = tb_tool.getcolshapestring(col)[0]
        # Convert `shape_string` into a tuple that numpy understands
        extra_dimensions = tuple(
            [
                int(idx)
                for idx in shape_string.replace("[", "").replace("]", "").split(", ")
            ]
        )
    except RuntimeError:
        extra_dimensions = ()

    #################################################################################

    # Get dtype of the column. Only read first row from disk
    col_dtype = np.array(tb_tool.col(col)[0]).dtype
    # Use a custom/safe fill value (https://github.com/casangi/xradio/issues/219)
    fill_value = get_pad_value(col_dtype)

    # Construct a numpy array to populate. `data` has shape (n_times, n_baselines, n_frequencies, n_polarizations)
    data = np.full(cshape + extra_dimensions, fill_value, dtype=col_dtype)

    # Use built-in casacore table iterator to populate the data column by unique times.
    if use_table_iter:
        start_row = 0
        for ts in tb_tool.iter("TIME", sort=False):
            num_rows = ts.nrows()

            # Create small temporary array to store the partial column
            tmp_arr = np.full(
                (num_rows,) + extra_dimensions, fill_value, dtype=col_dtype
            )

            # Note we don't use `getcol()` because it's less safe. See:
            # https://github.com/casacore/python-casacore/issues/130#issuecomment-463202373
            ts.getcolnp(col, tmp_arr)

            # Get the slice of rows contained in `tmp_arr`.
            # Used to get the relevant integer indexes from `tidxs` and `bidxs`
            tmp_slice = slice(start_row, start_row + num_rows)

            # Copy `tmp_arr` into correct elements of `tmp_arr`
            data[tidxs[tmp_slice], bidxs[tmp_slice]] = tmp_arr
            start_row += num_rows
    else:
        data[tidxs, bidxs] = tb_tool.getcol(col)

    return data



from importlib.metadata import version
import toolviper.utils.logger as logger, multiprocessing, psutil
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import xarray as xr

from .cds import CASAVisSet
from .stokes_types import stokes_types
from xradio._utils.list_and_array import get_pad_value


def make_coords(
    xds: xr.Dataset, ddi: int, subtables: Tuple[xr.Dataset, ...]
) -> Dict[str, np.ndarray]:
    """
    Make the coords to be added to a partition or chunk (besides
    the time, baseline) basic structure

    Grabs:
    - channel (center) frequency values from the spw subtable
    - pol idxs from the pol+ddi subtables -> pol names via the stokes_types
    - antenna IDs from antenna subtable

    Parameters
    ----------
    xds : xr.Dataset

    ddi : int

    subtables: Tuple[xr.Dataset, ...]


    Returns
    -------
    Dict[str, np.ndarray]
    """
    ant_xds, ddi_xds, spw_xds, pol_xds = subtables
    freq = spw_xds.CHAN_FREQ.values[
        ddi_xds.SPECTRAL_WINDOW_ID.values[ddi], : xds.freq.shape[0]
    ]
    pol_ids = pol_xds.CORR_TYPE.values[
        ddi_xds.POLARIZATION_ID.values[ddi], : xds.pol.shape[0]
    ]
    pol_names = np.vectorize(stokes_types.get)(pol_ids)
    ant_id = ant_xds.antenna_id.values
    coords = {
        "freq": freq,
        "pol": pol_names,
        "antenna_id": ant_id,
        # These will be metainfo in partitions
        # "spw_id": [ddi_xds["spectral_window_id"].values[ddi]],
        # "pol_id": [ddi_xds["polarization_id"].values[ddi]],
    }
    return coords


def vis_xds_packager_cds(
    subtables: List[Tuple[str, xr.Dataset]],
    partitions: Dict[Any, xr.Dataset],
    descr_add: str = "",
) -> CASAVisSet:
    """
    Takes a a list of subtable xds datasets and a dictionary of data
    partition xds datasets and and packages them as a CASA vis dataset
    (cds)

    Parameters
    ----------
    partitions : List[Tuple[str, xr.Dataset]]
        data partiions as xds datasets
    subtables : Dict[Any, xr.Dataset]
        subtables as xds datasets
    descr_add : str (Default value = "")
        substring to add to the short descr string of the cds

    Returns
    -------
    CASAVisSet
        A "cds" - container for the metainfo subtables and data partitions
    """
    vers = version("xradio")

    return CASAVisSet(
        subtables,
        partitions,
        f"CASA vis set produced by xradio {vers}/{descr_add}",
    )


def vis_xds_packager_mxds(
    partitions: Dict[Any, xr.Dataset],
    subtables: List[Tuple[str, xr.Dataset]],
    add_global_coords: bool = True,
) -> xr.Dataset:
    """
    Takes a dictionary of data partition xds datasets and a list of
    subtable xds datasets and packages them as a dataset of datasets
    (mxds)

    Parameters
    ----------
    partitions : Dict[Any, xr.Dataset]
        data partiions as xds datasets
    subtables : List[Tuple[str, xr.Dataset]]
        subtables as xds datasets
        :add_global_coords: whether to add coords to the output mxds
    add_global_coords: bool (Default value = True)

    Returns
    -------
    xr.Dataset
        A "mxds" - xr.dataset of datasets
    """
    mxds = xr.Dataset(attrs={"metainfo": subtables, "partitions": partitions})

    if add_global_coords:
        mxds = mxds.assign_coords(make_global_coords(mxds))

    return mxds


def make_global_coords(mxds: xr.Dataset) -> Dict[str, xr.DataArray]:
    coords = {}
    metainfo = mxds.attrs["metainfo"]
    if "antenna" in metainfo:
        coords["antenna_ids"] = metainfo["antenna"].antenna_id.values
        coords["antennas"] = xr.DataArray(
            metainfo["antenna"].NAME.values, dims=["antenna_ids"]
        )
    if "field" in metainfo:
        coords["field_ids"] = metainfo["field"].field_id.values
        coords["fields"] = xr.DataArray(
            metainfo["field"].NAME.values, dims=["field_ids"]
        )
    if "feed" in mxds.attrs:
        coords["feed_ids"] = metainfo["feed"].FEED_ID.values
    if "observation" in metainfo:
        coords["observation_ids"] = metainfo["observation"].observation_id.values
        coords["observations"] = xr.DataArray(
            metainfo["observation"].PROJECT.values, dims=["observation_ids"]
        )
    if "polarization" in metainfo:
        coords["polarization_ids"] = metainfo["polarization"].pol_setup_id.values
    if "source" in metainfo:
        coords["source_ids"] = metainfo["source"].SOURCE_ID.values
        coords["sources"] = xr.DataArray(
            metainfo["source"].NAME.values, dims=["source_ids"]
        )
    if "spectral_window" in metainfo:
        coords["spw_ids"] = metainfo["spectral_window"].spw_id.values
    if "state" in metainfo:
        coords["state_ids"] = metainfo["state"].STATE_ID.values

    return coords


def expand_xds(xds: xr.Dataset) -> xr.Dataset:
    """
    expand single (row) dimension of xds to (time, baseline)

    Parameters
    ----------
    xds : xr.Dataset
        "flat" dataset (with row dimension - without (time, baseline) dimensions)

    Returns
    -------
    xr.Dataset
        expanded dataset, with (time, baseline) dimensions
    """
    assert "baseline" not in xds.coords

    txds = xds.copy()

    unique_baselines, baselines = np.unique(
        [txds.baseline_ant1_id.values, txds.baseline_ant2_id.values],
        axis=1,
        return_inverse=True,
    )

    txds["baseline"] = xr.DataArray(baselines.astype("int32"), dims=["row"])

    try:
        txds = (
            txds.set_index(row=["time", "baseline"])
            .unstack("row")
            .transpose("time", "baseline", ...)
        )
        # unstack changes type to float when it needs to introduce NaNs, so
        # we need to reset to the proper type. Avoid if possible, as the
        # astype are costly
        for dv in txds.data_vars:
            if txds[dv].dtype != xds[dv].dtype:
                txds[dv] = txds[dv].astype(xds[dv].dtype)
    except Exception as exc:
        logger.warning(
            f"WARNING: Cannot expand rows to (time, baseline), "
            f"possibly duplicate values in (time, baseline). Exception: {exc}."
            f"\nDataset: {txds=}"
        )
        txds = xds.copy()

    return txds


def flatten_xds(xds: xr.Dataset) -> xr.Dataset:
    """
    flatten (time, baseline) dimensions of xds back to single dimension (row)

    Parameters
    ----------
    xds : xr.Dataset


    Returns
    -------
    xr.Dataset
        Dataset in flat form (back to 'row' dimension as read by casacore tables)
    """
    txds = xds.copy()

    # flatten the time x baseline dimensions of main table
    if ("time" in xds.sizes) and ("baseline" in xds.sizes):
        txds = xds.stack({"row": ("time", "baseline")}).transpose("row", ...)
        # compute for issue https://github.com/hainegroup/oceanspy/issues/332
        # drop=True silently does compute (or at least used to)

        fill_value_int32 = get_pad_value(np.int32)
        txds = txds.where(
            (
                (txds.STATE_ID != fill_value_int32)
                & (txds.FIELD_ID != fill_value_int32)
            ).compute(),
            drop=True,
        )  # .unify_chunks()

        # re-assigning (implicitly dropping index coords) one by one produces
        # DeprecationWarnings: https://github.com/pydata/xarray/issues/6505
        astyped_data_vars = dict(xds.data_vars)
        for dv in list(txds.data_vars):
            if txds[dv].dtype != xds[dv].dtype:
                astyped_data_vars[dv] = txds[dv].astype(xds[dv].dtype)
            else:
                astyped_data_vars[dv] = txds[dv]

        flat_xds = xr.Dataset(astyped_data_vars, coords=txds.coords, attrs=txds.attrs)
        flat_xds = flat_xds.reset_index(["time", "baseline"])

    else:
        flat_xds = txds

    return flat_xds


####################################
# xautomatically compute best data chunking
def optimal_chunking(
    ndim: Union[int, None] = None,
    didxs: Union[Tuple[int], List[int], None] = None,
    chunk_size: str = "auto",
    data_shape: Union[tuple, None] = None,
) -> tuple:
    """
    Determine the optimal chunk shape for reading an MS or Image based
    on machine resources and intended operations

    Parameters
    ----------
    ndim : Union[int, None] = None
        number of dimensions to chunk. An MS is 3, an
        expanded MS is 4. An image could be anywhere from 2 to 5. Not
        needed if data_shape is given.
    didxs : Union[Tuple[int], List[int], None] = None
        dimension indices over which subsequent operations
        will be performed. Values should be less than ndim. Tries to
        reduce inter-process communication of data contents. Needs to
        know the shape to do this well. Default None balances chunk size
        across all dimensions.
    chunk_size : str (Default value = "auto")
        target chunk size ('large', 'small', 'auto').
        Default 'auto' tries to guess by looking at CPU core count and
        available memory.
    data_shape : Union[tuple, None] = None
        shape of the total MS DDI or Image data. Helps
        to know. Default None does not optimize based on shape

    Returns
    -------
    tuple
        optimal chunking for reading the ms (row, chan, pol)
    """
    assert (ndim is not None) or (
        data_shape is not None
    ), "either ndim or data_shape must be given"
    assert chunk_size in ["large", "small", "auto"], "invalid chunk_size parameter"
    if ndim is None:
        ndim = len(data_shape)

    opt_dims = (
        didxs if (didxs is not None) and (len(didxs) > 0) else np.arange(ndim)
    )  # maximize these dim chunk sizes
    nonopt_dims = np.setdiff1d(np.arange(ndim), opt_dims)  # at the expense of these

    max_chunk_sizes = (
        data_shape
        if data_shape is not None
        else [dd for ii, dd in enumerate([10000, 10000, 10000, 4, 10]) if ii < ndim]
    )
    min_chunk_sizes = (
        np.ceil(np.array(data_shape) / 80).astype(int)
        if data_shape is not None
        else (
            [1000, 1, 1]
            if ndim == 3
            else [dd for ii, dd in enumerate([10, 10, 1, 1, 1]) if ii < ndim]
        )
    )
    target_size = 175 * 1024**2 / 8  # ~175 MB chunk worst case with 8-byte DATA column
    bytes_per_core = int(
        round(
            ((psutil.virtual_memory().available * 0.10) / multiprocessing.cpu_count())
        )
    )
    if data_shape is not None:
        bytes_per_core = min(
            bytes_per_core, np.prod(data_shape) * 8 / 2
        )  # ensure at least two chunks
    if chunk_size == "large":
        target_size = target_size * 6  # ~1 GB
    if chunk_size == "auto":
        target_size = max(min(target_size * 6, bytes_per_core / 8), target_size)

    # start by setting the optimized dims to their max size and non-optimized dims to their min size
    chunks = np.zeros((ndim), dtype="int")
    chunks[opt_dims] = np.array(max_chunk_sizes)[opt_dims]
    chunks[nonopt_dims] = np.array(min_chunk_sizes)[nonopt_dims]

    # iteratively walk towards an optimal chunk size
    # iteration is needed because rounding to nearest integer index can make a big different (2x) in chunk size
    # for small dimensions like pol
    for ii in range(10):
        # if the resulting size is too big, reduce the sizes of the optimized dimensions
        if (np.prod(chunks) > target_size) and (len(opt_dims) > 0):
            chunks[opt_dims] = np.round(
                chunks[opt_dims]
                * (target_size / np.prod(chunks)) ** (1 / len(opt_dims))
            )
        # else if the resulting size is too small, increase the sizes of the non-optimized dimensions
        elif (np.prod(chunks) < target_size) and (len(nonopt_dims) > 0):
            chunks[nonopt_dims] = np.round(
                chunks[nonopt_dims]
                * (target_size / np.prod(chunks)) ** (1 / len(nonopt_dims))
            )
        chunks = np.min((chunks, max_chunk_sizes), axis=0)
        chunks = np.max((chunks, min_chunk_sizes), axis=0)

    return tuple(chunks)


def calc_optimal_ms_chunk_shape(
    memory_available_in_bytes, shape, element_size_in_bytes, column_name
) -> int:
    """
    Calculates the max number of rows (1st dim in shape) of a variable
    that can be fit in the memory for a thread.

    Parameters
    ----------
    memory_available_in_bytes :

    shape :

    element_size_in_bytes :

    column_name :


    Returns
    -------
    int
    """
    factor = 0.8  # Account for memory used by other objects in thread.
    # total_mem = np.prod(shape)*element_size_in_bytes
    single_row_mem = np.prod(shape[1:]) * element_size_in_bytes

    if not single_row_mem < factor * memory_available_in_bytes:
        msg = (
            "Not engough memory in a thread to contain a row of "
            f"{column_name}. Need at least {single_row_mem / factor}"
            " bytes."
        )
        raise RuntimeError(msg)

    rows_chunk_size = int((factor * memory_available_in_bytes) / single_row_mem)

    if rows_chunk_size > shape[0]:
        rows_chunk_size = shape[0]

    logger.debug(
        "Numbers of rows in chunk for " + column_name + ": " + str(rows_chunk_size)
    )

    return rows_chunk_size



from typing import Dict, TypedDict, Union

import xarray as xr


PartitionIds = TypedDict(
    "PartitionIds",
    {
        "array_id": int,
        "observation_id": int,
        "pol_setup_id": int,
        "processor_id": int,
        "spw_id": int,
    },
)

VisGroup = TypedDict(
    "VisGroup",
    {
        "seq_id": int,
        "vis": str,
        "flag": str,
        "weight": str,
        "uvw": str,
        "imaging_weight": Union[str, None],
        "descr": str,
    },
)


def make_vis_group_attr(xds: xr.Dataset) -> Dict:
    """
    Add an attribute with the initial data/vis groups that have been
    read from the MS (DATA / CORRECTED_DATA / MODEL_DATA)

    Parameters
    ----------
    xds : xr.Dataset
        dataset to make the vis_group depending on its data_vars

    Returns
    -------
    Dict
        vis_group derived form this dataset
    """
    msv2_extended_vis_vars = ["vis", "vis_corrected", "vis_model"]
    msv2_col_names = ["DATA", "CORRECTED_DATA", "MODEL_DATA"]
    # example test MS with imaging_weight col: vla/ic2233_1.ms
    imgw_name = "imaging_weight"
    img_weight = imgw_name if imgw_name in xds.data_vars else None
    vis_groups = {}
    seq_id = 1
    for var, msv2_name in zip(msv2_extended_vis_vars, msv2_col_names):
        if var in xds.data_vars:
            grp: VisGroup = {
                "seq_id": seq_id,
                "vis": var,
                "flag": "flag",
                "weight": "weight",
                "uvw": "uvw",
                "imaging_weight": img_weight,
                "descr": "xradio.vis.ms.read_ms from MSv2",
            }
            seq_id += 1
            vis_groups[f"MeasurementSet/{msv2_name}"] = grp

    return vis_groups


def init_partition_ids(
    xds: xr.Dataset,
    ddi: int,
    ddi_xds: xr.Dataset,
    part_ids: PartitionIds,
) -> PartitionIds:
    spw_id = ddi_xds.SPECTRAL_WINDOW_ID.values[ddi]
    pol_setup_id = ddi_xds.POLARIZATION_ID.values[ddi]
    ids: PartitionIds = {
        # The -1 are expected to be be updated from part_ids
        "array_id": -1,
        "observation_id": -1,
        "pol_setup_id": pol_setup_id,
        "processor_id": -1,
        "spw_id": spw_id,
    }
    ids.update(part_ids)

    return ids


def add_partition_attrs(
    xds: xr.Dataset,
    ddi: int,
    ddi_xds: xr.Dataset,
    part_ids: PartitionIds,
    other_attrs: Dict,
) -> xr.Dataset:
    """
    add attributes to the xr.Dataset:
    - sub-dict of partition-id related ones
    - sub-dict of data/vis groups
    - sub-dict of attributes coming from the lower level read
      functions (MSv2 stuff, etc.)

    Produces the partition IDs that can be retrieved from the DD subtable and also
    adds the ones passed in part_ids

    Parameters
    ----------
    xds : xr.Dataset
        dataset partition
    ddi : int
        DDI of this partition
    ddi_xds : xr.Dataset
        dataset for the DATA_DESCRIPTION subtable
    part_ids : PartitionIds
        partition id attrs
    other_attrs : Dict
        additional attributes produced by the read functions

    Returns
    -------
    xr.Dataset
        dataset with attributes added
    """

    xds = xds.assign_attrs(
        {"partition_ids": init_partition_ids(xds, ddi, ddi_xds, part_ids)}
    )
    xds = xds.assign_attrs({"vis_groups": make_vis_group_attr(xds)})
    xds = xds.assign_attrs(other_attrs)
    return xds



from typing import Generator
from contextlib import contextmanager

from casacore import tables


@contextmanager
def open_table_ro(infile: str) -> Generator[tables.table, None, None]:
    table = tables.table(
        infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )
    try:
        yield table
    finally:
        table.close()


@contextmanager
def open_query(table: tables.table, query: str) -> Generator[tables.table, None, None]:
    ttq = tables.taql(query)
    try:
        yield ttq
    finally:
        ttq.close()



from casacore import tables
from contextlib import contextmanager
from typing import Dict, Generator

# common casacore table handling code


def extract_table_attributes(infile: str) -> Dict[str, Dict]:
    """
    return a dictionary of table attributes created from MS keywords and column descriptions
    """
    with open_table_ro(infile) as tb_tool:
        kwd = tb_tool.getkeywords()
        attrs = dict([(kk, kwd[kk]) for kk in kwd if kk not in os.listdir(infile)])
        cols = tb_tool.colnames()
        column_descriptions = {}
        for col in cols:
            column_descriptions[col] = tb_tool.getcoldesc(col)
        attrs["column_descriptions"] = column_descriptions
        attrs["info"] = tb_tool.info()
    return attrs


@contextmanager
def open_table_ro(infile: str) -> Generator[tables.table, None, None]:
    table = tables.table(
        infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )
    try:
        yield table
    finally:
        table.close()


@contextmanager
def open_table_rw(outfile: str) -> Generator[tables.table, None, None]:
    table = tables.table(
        outfile, readonly=False, lockoptions={"option": "permanentwait"}, ack=False
    )
    try:
        yield table
    finally:
        table.close()



import os
from xradio.measurement_set import ProcessingSet
from typing import Dict, Union


def load_processing_set(
    ps_store: str,
    sel_parms: dict,
    data_variables: Union[list, None] = None,
    load_sub_datasets: bool = True,
) -> ProcessingSet:
    """Loads a processing set into memory.

    Parameters
    ----------
    ps_store : str
        String of the path and name of the processing set. For example '/users/user_1/uid___A002_Xf07bba_Xbe5c_target.lsrk.vis.zarr' for a file stored on a local file system, or 's3://viper-test-data/Antennae_North.cal.lsrk.split.vis.zarr/' for a file in AWS object storage.
    sel_parms : dict
        A dictionary where the keys are the names of the ms_xds's and the values are slice_dicts.
        slice_dicts: A dictionary where the keys are the dimension names and the values are slices.
        For example::

            {
                'ms_v4_name_1': {'frequency': slice(0, 160, None),'time':slice(0,100)},
                ...
                'ms_v4_name_n': {'frequency': slice(0, 160, None),'time':slice(0,100)},
            }

    data_variables : Union[list, None], optional
        The list of data variables to load into memory for example ['VISIBILITY', 'WEIGHT, 'FLAGS']. By default None which will load all data variables into memory.
    load_sub_datasets : bool, optional
        If true sub-datasets (for example weather_xds, antenna_xds, pointing_xds, system_calibration_xds ...) will be loaded into memory, by default True.

    Returns
    -------
    ProcessingSet
        In memory representation of processing set (data is represented by Dask.arrays).
    """
    from xradio._utils.zarr.common import _open_dataset, _get_file_system_and_items

    file_system, ms_store_list = _get_file_system_and_items(ps_store)

    ps = ProcessingSet()
    for ms_name, ms_xds_isel in sel_parms.items():
        ms_store = os.path.join(ps_store, ms_name)
        correlated_store = os.path.join(ms_store, "correlated_xds")

        xds = _open_dataset(
            correlated_store,
            file_system,
            ms_xds_isel,
            data_variables,
            load=True,
        )
        data_groups = xds.attrs["data_groups"]

        if load_sub_datasets:
            from xradio.measurement_set.open_processing_set import _open_sub_xds

            sub_xds_dict, field_and_source_xds_dict = _open_sub_xds(
                ms_store, file_system=file_system, load=True, data_groups=data_groups
            )

            xds.attrs = {
                **xds.attrs,
                **sub_xds_dict,
            }
            for data_group_name, data_group_vals in data_groups.items():

                xds[data_group_vals["correlated_data"]].attrs[
                    "field_and_source_xds"
                ] = field_and_source_xds_dict[data_group_name]

        ps[ms_name] = xds

    return ps


class ProcessingSetIterator:
    def __init__(
        self,
        sel_parms: dict,
        input_data_store: str,
        input_data: Union[Dict, ProcessingSet, None] = None,
        data_variables: list = None,
        load_sub_datasets: bool = True,
    ):
        """An iterator that will go through a processing set one MS v4 at a time.

        Parameters
        ----------
        sel_parms : dict
            A dictionary where the keys are the names of the ms_xds's and the values are slice_dicts.
            slice_dicts: A dictionary where the keys are the dimension names and the values are slices.
            For example::

                {
                    'ms_v4_name_1': {'frequency': slice(0, 160, None),'time':slice(0,100)},
                    ...
                    'ms_v4_name_n': {'frequency': slice(0, 160, None),'time':slice(0,100)},
                }
        input_data_store : str
            String of the path and name of the processing set. For example '/users/user_1/uid___A002_Xf07bba_Xbe5c_target.lsrk.vis.zarr'.
        input_data : Union[Dict, processing_set, None], optional
            If the processing set is in memory already it can be supplied here. By default None which will make the iterator load data using the supplied input_data_store.
        data_variables : list, optional
            The list of data variables to load into memory for example ['VISIBILITY', 'WEIGHT, 'FLAGS']. By default None which will load all data variables into memory.
        load_sub_datasets : bool, optional
            If true sub-datasets (for example weather_xds, antenna_xds, pointing_xds, system_calibration_xds ...) will be loaded into memory, by default True.
        """

        self.input_data = input_data
        self.input_data_store = input_data_store
        self.sel_parms = sel_parms
        self.xds_name_iter = iter(sel_parms.keys())
        self.data_variables = data_variables
        self.load_sub_datasets = load_sub_datasets

    def __iter__(self):
        return self

    def __next__(self):
        try:
            xds_name = next(self.xds_name_iter)
        except Exception as e:
            raise StopIteration

        if self.input_data is None:
            slice_description = self.sel_parms[xds_name]
            ps = load_processing_set(
                ps_store=self.input_data_store,
                sel_parms={xds_name: slice_description},
                data_variables=self.data_variables,
                load_sub_datasets=self.load_sub_datasets,
            )
            xds = ps.get(0)
        else:
            xds = self.input_data[xds_name]  # In memory

        return xds



import os
from toolviper.utils.logger import setup_logger

# _logger_name = "xradio"
# if os.getenv("VIPER_LOGGER_NAME") != _logger_name:
#     os.environ["VIPER_LOGGER_NAME"] = _logger_name
#     setup_logger(
#         logger_name="xradio",
#         log_to_term=True,
#         log_to_file=False,  # True
#         log_file="xradio-logfile",
#         log_level="DEBUG",
#     )



import numpy as np


_np_types = {
    "complex128": np.complex128,
    "complex64": np.complex64,
    "float64": np.float64,
    "float16": np.float16,
    "float32": np.float32,
    "int8": np.int8,
    "int16": np.int16,
    "int32": np.int32,
    "int64": np.int64,
}

_top_level_sub_xds = "_attrs_xds"



from casacore import images
from contextlib import contextmanager
import numpy as np
from typing import Dict, Generator, List, Union


@contextmanager
def _open_image_ro(infile: str) -> Generator[images.image, None, None]:
    image = images.image(infile)
    try:
        yield image
    finally:
        # there is no obvious way to close a python-casacore image, so
        # just delete the object to clear it from the table cache
        del image


"""
@contextmanager
def _open_image_rw(
    infile: str, mask: str, shape: tuple
) -> Generator[images.image, None, None]:
    image = images.image(infile, maskname=mask, shape=shape)
    try:
        yield image
    finally:
        del image
"""


@contextmanager
def _create_new_image(
    outfile: str, shape: List[int], mask="", value="default"
) -> Generator[images.image, None, None]:
    # new image will be opened rw
    # the crux of the issue here seems to be that python has no single
    # precision floating point value, but we need single precision for
    # most images. The image constructor gets that right if values isn't
    # suppliked. Hence two calls with values not present and values present
    if value == "default":
        image = images.image(outfile, maskname=mask, shape=shape)
    else:
        image = images.image(outfile, maskname=mask, shape=shape, values=value)
    try:
        yield image
    finally:
        del image


_active_mask = "active_mask"
# _native_types = ["FREQ", "VRAD", "VOPT", "BETA", "WAVE", "AWAV"]
_object_name = "object_name"
_pointing_center = "pointing_center"



import toolviper.utils.logger as logger
from typing import Any, Dict, List, Tuple, Union

import dask, dask.array
import numpy as np
import xarray as xr
import pandas as pd

from casacore import tables

from .read import (
    read_flat_col_chunk,
    read_col_chunk,
    convert_casacore_time,
    extract_table_attributes,
    add_units_measures,
)

from .table_query import open_table_ro, open_query
from xradio._utils.list_and_array import (
    unique_1d,
    pairing_function,
    inverse_pairing_function,
)

rename_msv2_cols = {
    "ANTENNA1": "antenna1_id",
    "ANTENNA2": "antenna2_id",
    "FEED1": "feed1_id",
    "FEED2": "feed2_id",
    # optional cols:
    "WEIGHT_SPECTRUM": "WEIGHT",
    "CORRECTED_DATA": "VIS_CORRECTED",
    "DATA": "VIS",
    "MODEL_DATA": "VIS_MODEL",
    "FLOAT_DATA": "AUTOCORR",
}


def rename_vars(mvars: Dict[str, xr.DataArray]) -> Dict[str, xr.DataArray]:
    """
    Apply rename rules. Also preserve ordering of data_vars

    Note: not using xr.DataArray.rename because we have optional
    column renames and rename complains if some of the names passed
    are not present in the dataset

    Parameters
    ----------
    mvars : Dict[str, xr.DataArray]
        dictionary of data_vars to be used to create an xr.Dataset

    Returns
    -------
    Dict[str, xr.DataArray]
        similar dictionary after applying MSv2 => MSv3/ngCASA renaming rules
    """
    renamed = {
        rename_msv2_cols[name] if name in rename_msv2_cols else name: var
        for name, var in mvars.items()
    }

    return renamed


def redim_id_data_vars(mvars: Dict[str, xr.DataArray]) -> Dict[str, xr.DataArray]:
    """
    Changes:
    Several id data variables to drop its baseline dim
    The antenna id data vars:
     From MS (antenna1_id(time, baseline), antenna2_id(time,baseline)
     To cds (baseline_ant1_id(baseline), baseline_ant2_id(baseline)

    Parameters
    ----------
    mvars : Dict[str, xr.DataArray]
        data variables being prepared for a partition xds

    Returns
    -------
    Dict[str, xr.DataArray]
        data variables with the ant id ones modified to cds type
    """
    # Vars to drop baseline dim
    var_names = [
        "ARRAY_ID",
        "OBSERVATION_ID",
        "PROCESSOR_ID",
        "SCAN_NUMBER",
        "STATE_ID",
    ]
    for vname in var_names:
        if "baseline" in mvars[vname].coords:
            mvars[vname] = mvars[vname].sel(baseline=0, drop=True)

    for idx in ["1", "2"]:
        new_name = f"baseline_ant{idx}_id"
        mvars[new_name] = mvars.pop(f"antenna{idx}_id")
        if "time" in mvars[new_name].coords:
            mvars[new_name] = mvars[new_name].sel(time=0, drop=True)

    return mvars


def get_partition_ids(mtable: tables.table, taql_where: str) -> Dict:
    """
    Get some of the partition IDs that we have to retrieve from some
    of the top level ID/sorting cols of the main table of the MS.

    Parameters
    ----------
    mtable : tables.table
        MS main table
    taql_where : str
        where part that defines the partition in TaQL

    Returns
    -------
    Dict
        ids of array, observation, and processor
    """

    taql_ids = f"select DISTINCT ARRAY_ID, OBSERVATION_ID, PROCESSOR_ID from $mtable {taql_where}"
    with open_query(mtable, taql_ids) as query:
        # array_id, observation_id, processor_id
        array_id = unique_1d(query.getcol("ARRAY_ID"))
        obs_id = unique_1d(query.getcol("OBSERVATION_ID"))
        proc_id = unique_1d(query.getcol("PROCESSOR_ID"))
        check_vars = [
            (array_id, "array_id"),
            (obs_id, "observation_id"),
            (proc_id, "processor_id"),
        ]
        for var, var_name in check_vars:
            if len(var) != 1:
                logger.warning(
                    f"Did not get exactly one {var_name} (got {var} for this partition. TaQL: {taql_where}"
                )

    pids = {
        "array_id": list(array_id),
        "observation_id": list(obs_id),
        "processor_id": list(proc_id),
    }
    return pids


def read_expanded_main_table(
    infile: str,
    ddi: int = 0,
    scan_state: Union[Tuple[int, int], None] = None,
    ignore_msv2_cols: Union[list, None] = None,
    chunks: Tuple[int, ...] = (400, 200, 100, 2),
) -> Tuple[xr.Dataset, Dict[str, Any], Dict[str, Any]]:
    """
    Reads one partition from the main table, all columns.
    This is the expanded version (time, baseline) dims.

    Chunk tuple: (time, baseline, freq, pol)

    Parameters
    ----------
    infile : str

    ddi : int  (Default value = 0)

    scan_state : Union[Tuple[int, int], None] (Default value = None)

    ignore_msv2_cols: Union[list, None] (Default value = None)

    chunks: Tuple[int, ...] (Default value = (400, 200, 100, 2))


    Returns
    -------
    Tuple[xr.Dataset, Dict[str, Any], Dict[str, Any]]
    """
    if ignore_msv2_cols is None:
        ignore_msv2_cols = []

    taql_where = f"where DATA_DESC_ID = {ddi}"
    if scan_state:
        # get partitions by scan/state
        scan, state = scan_state
        if type(state) == np.ndarray:
            state_ids_or = " OR STATE_ID = ".join(np.char.mod("%d", state))
            taql_where += f" AND (STATE_ID = {state_ids_or})"
        elif state:
            taql_where += f" AND SCAN_NUMBER = {scan} AND STATE_ID = {state}"
        elif scan:
            # scan can also be None, when partition_scheme='intent'
            # but the STATE table is empty!
            taql_where += f" AND SCAN_NUMBER = {scan}"

    with open_table_ro(infile) as mtable:
        # one partition, select just the specified ddi (+ scan/subscan)
        taql_main = f"select * from $mtable {taql_where}"
        with open_query(mtable, taql_main) as tb_tool:
            if tb_tool.nrows() == 0:
                tb_tool.close()
                mtable.close()
                return xr.Dataset(), {}, {}

            xds, attrs = read_main_table_chunks(
                infile, tb_tool, taql_where, ignore_msv2_cols, chunks
            )
            part_ids = get_partition_ids(tb_tool, taql_where)

    return xds, part_ids, attrs


def read_main_table_chunks(
    infile: str,
    tb_tool: tables.table,
    taql_where: str,
    ignore_msv2_cols: Union[list, None] = None,
    chunks: Tuple[int, ...] = (400, 200, 100, 2),
) -> Tuple[xr.Dataset, Dict[str, Any]]:
    """
    Iterates through the time,baseline chunks and reads slices from
    all the data columns.

    Parameters
    ----------
    infile : str

    tb_tool : tables.table

    taql_where : str

    ignore_msv2_cols : Union[list, None] (Default value = None)

    chunks: Tuple[int, ...] (Default value = (400, 200, 100, 2))


    Returns
    -------
    Tuple[xr.Dataset, Dict[str, Any]]
    """
    baselines = get_baselines(tb_tool)

    col_names = tb_tool.colnames()
    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in col_names
        if tb_tool.iscelldefined(col, 0)
    ]
    chan_cnt, pol_cnt = [(cc[0], cc[1]) for cc in cshapes if len(cc) == 2][0]

    unique_times, tol = get_utimes_tol(tb_tool, taql_where)

    tvars = {}
    n_baselines = len(baselines)
    n_unique_times = len(unique_times)
    n_time_chunks = chunks[0]
    n_baseline_chunks = chunks[1]
    # loop over time chunks
    for time_chunk in range(0, n_unique_times, n_time_chunks):
        time_start = unique_times[time_chunk] - tol
        time_end = (
            unique_times[min(n_unique_times, time_chunk + n_time_chunks) - 1] + tol
        )

        # chunk time length
        ctlen = min(n_unique_times, time_chunk + n_time_chunks) - time_chunk

        bvars = {}
        # loop over baseline chunks
        for baseline_chunk in range(0, n_baselines, n_baseline_chunks):
            cblen = min(n_baselines - baseline_chunk, n_baseline_chunks)

            # read the specified chunk of data
            # def read_chunk(infile, ddi, times, blines, chans, pols):
            ttql = f"TIME BETWEEN {time_start} and {time_end}"
            ant1_start = baselines[baseline_chunk][0]
            ant1_end = baselines[cblen + baseline_chunk - 1][0]
            atql = f"ANTENNA1 BETWEEN {ant1_start} and {ant1_end}"
            ts_taql = f"select * from $mtable {taql_where} AND {ttql} AND {atql}"
            with open_query(None, ts_taql) as query_times_ants:
                tidxs = (
                    np.searchsorted(
                        unique_times, query_times_ants.getcol("TIME", 0, -1)
                    )
                    - time_chunk
                )
                ts_ant1, ts_ant2 = (
                    query_times_ants.getcol("ANTENNA1", 0, -1),
                    query_times_ants.getcol("ANTENNA2", 0, -1),
                )

            ts_bases = np.column_stack((ts_ant1, ts_ant2))

            bidxs = get_baseline_indices(baselines, ts_bases) - baseline_chunk

            # some antenna 2's will be out of bounds for this chunk, store rows that are in bounds
            didxs = np.where(
                (bidxs >= 0) & (bidxs < min(chunks[1], n_baselines - baseline_chunk))
            )[0]

            delayed_params = (infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs)

            read_all_cols_bvars(
                tb_tool, chunks, chan_cnt, ignore_msv2_cols, delayed_params, bvars
            )

        concat_bvars_update_tvars(bvars, tvars)

    dims = ["time", "baseline", "freq", "pol"]
    mvars = concat_tvars_to_mvars(dims, tvars, pol_cnt, chan_cnt)

    mcoords = {
        "time": xr.DataArray(convert_casacore_time(unique_times), dims=["time"]),
        "baseline": xr.DataArray(np.arange(n_baselines), dims=["baseline"]),
    }

    # add xds global attributes
    cc_attrs = extract_table_attributes(infile)
    attrs = {"other": {"msv2": {"ctds_attrs": cc_attrs, "bad_cols": ignore_msv2_cols}}}
    # add per data var attributes
    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    mvars = rename_vars(mvars)
    mvars = redim_id_data_vars(mvars)
    xds = xr.Dataset(mvars, coords=mcoords)

    return xds, attrs


def get_utimes_tol(mtable: tables.table, taql_where: str) -> Tuple[np.ndarray, float]:
    taql_utimes = f"select DISTINCT TIME from $mtable {taql_where}"
    with open_query(mtable, taql_utimes) as query_utimes:
        utimes = unique_1d(query_utimes.getcol("TIME", 0, -1))
        # add a tol around the time ranges returned by taql
        if len(utimes) < 2:
            tol = 1e-5
        else:
            tol = np.diff(utimes).min() / 4

    return utimes, tol


def get_baselines(tb_tool: tables.table) -> np.ndarray:
    """
    Gets the unique baselines from antenna 1 and antenna 2 ids.

    Uses a pairing function and inverse pairing function to decrease the
    computation time of finding unique values.

    Parameters
    ----------
    tb_tool : tables.table
        MeasurementSet table to get the antenna ids.

    Returns
    -------
    unique_baselines : np.ndarray
        a 2D array of unique antenna pairs
        (baselines) from the MeasurementSet table provided.
    """
    ant1, ant2 = tb_tool.getcol("ANTENNA1", 0, -1), tb_tool.getcol("ANTENNA2", 0, -1)

    baselines = np.column_stack((ant1, ant2))

    # Using pairing function to reduce the computation time of finding unique values.
    baselines_paired = pairing_function(baselines)
    unique_baselines_paired = pd.unique(baselines_paired)
    unique_baselines = inverse_pairing_function(unique_baselines_paired)

    # Sorting the unique baselines.
    unique_baselines = unique_baselines[unique_baselines[:, 1].argsort()]
    unique_baselines = unique_baselines[
        unique_baselines[:, 0].argsort(kind="mergesort")
    ]

    return unique_baselines


def get_baseline_indices(
    unique_baselines: np.ndarray, baseline_set: np.ndarray
) -> np.ndarray:
    """
    Finds the baseline indices of a set of baselines using the unique baselines.

    Uses a pairing function to reduce the number of values so it's more
    efficient to find the indices.

    Parameters
    ----------
    unique_baselines : np.ndarray
        a 2D array of unique antenna pairs (baselines).
    baseline_set : np.ndarray
        a 2D array of antenna pairs (baselines). This array may contain duplicates.

    Returns
    -------
    baseline_indices : np.ndarray
        the indices of the baseline set that
        correspond to the unique baselines.
    """
    unique_baselines_paired = pairing_function(unique_baselines)
    baseline_set_paired = pairing_function(baseline_set)

    # Pairing function doesn't preserve order so they need to be sorted.
    unique_baselines_sorted = np.argsort(unique_baselines_paired)
    sorted_indices = np.searchsorted(
        unique_baselines_paired[unique_baselines_sorted],
        baseline_set_paired,
    )
    baseline_indices = unique_baselines_sorted[sorted_indices]

    return baseline_indices


def read_all_cols_bvars(
    tb_tool: tables.table,
    chunks: Tuple[int, ...],
    chan_cnt: int,
    ignore_msv2_cols: bool,
    delayed_params: Tuple,
    bvars: Dict[str, xr.DataArray],
) -> None:
    """
    Loops over each column and create delayed dask arrays

    Parameters
    ----------
    tb_tool : tables.table

    chunks : Tuple[int, ...]

    chan_cnt : int

    ignore_msv2_cols : bool

    delayed_params : Tuple

    bvars : Dict[str, xr.DataArray]


    Returns
    -------

    """

    col_names = tb_tool.colnames()
    for col in col_names:
        if (col in ignore_msv2_cols + ["TIME"]) or (not tb_tool.iscelldefined(col, 0)):
            continue
        if col not in bvars:
            bvars[col] = []

        cdata = tb_tool.getcol(col, 0, 1)[0]

        if len(cdata.shape) == 0:
            infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs = delayed_params
            cshape = (ctlen, cblen)
            delayed_col = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
            delayed_array = dask.delayed(read_col_chunk)(*delayed_col, None, None)
            bvars[col] += [dask.array.from_delayed(delayed_array, cshape, cdata.dtype)]

        elif col == "UVW":
            infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs = delayed_params
            cshape = (ctlen, cblen, 3)
            delayed_3 = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
            delayed_array = dask.delayed(read_col_chunk)(*delayed_3, None, None)
            bvars[col] += [dask.array.from_delayed(delayed_array, cshape, cdata.dtype)]

        elif len(cdata.shape) == 1:
            pol_list = []
            dd = 2 if cdata.shape == chan_cnt else 3
            for pc in range(0, cdata.shape[0], chunks[dd]):
                pols = (pc, min(cdata.shape[0], pc + chunks[dd]) - 1)
                infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs = delayed_params
                cshape = (
                    ctlen,
                    cblen,
                ) + (pols[1] - pols[0] + 1,)
                delayed_cs = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
                delayed_array = dask.delayed(read_col_chunk)(*delayed_cs, pols, None)
                pol_list += [
                    dask.array.from_delayed(delayed_array, cshape, cdata.dtype)
                ]
            bvars[col] += [dask.array.concatenate(pol_list, axis=2)]

        elif len(cdata.shape) == 2:
            chan_list = []
            for cc in range(0, cdata.shape[0], chunks[2]):
                chans = (cc, min(cdata.shape[0], cc + chunks[2]) - 1)
                pol_list = []
                for pc in range(0, cdata.shape[1], chunks[3]):
                    pols = (pc, min(cdata.shape[1], pc + chunks[3]) - 1)
                    (
                        infile,
                        ts_taql,
                        (ctlen, cblen),
                        tidxs,
                        bidxs,
                        didxs,
                    ) = delayed_params
                    cshape = (
                        ctlen,
                        cblen,
                    ) + (chans[1] - chans[0] + 1, pols[1] - pols[0] + 1)
                    delayed_cs = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
                    delayed_array = dask.delayed(read_col_chunk)(
                        *delayed_cs, chans, pols
                    )
                    pol_list += [
                        dask.array.from_delayed(delayed_array, cshape, cdata.dtype)
                    ]
                chan_list += [dask.array.concatenate(pol_list, axis=3)]
            bvars[col] += [dask.array.concatenate(chan_list, axis=2)]


def concat_bvars_update_tvars(
    bvars: Dict[str, xr.DataArray], tvars: Dict[str, xr.DataArray]
) -> None:
    """
    concats all the dask chunks from each baseline. This is intended to
    be called iteratively, for every time chunk iteration, once all the
    baseline chunks have been read.

    Parameters
    ----------
    bvars: Dict[str, xr.DataArray]

    tvars: Dict[str, xr.DataArray]


    Returns
    -------

    """
    for kk in bvars.keys():
        if len(bvars[kk]) == 0:
            continue
        if kk not in tvars:
            tvars[kk] = []
        tvars[kk] += [dask.array.concatenate(bvars[kk], axis=1)]


def concat_tvars_to_mvars(
    dims: List[str], tvars: Dict[str, xr.DataArray], pol_cnt: int, chan_cnt: int
) -> Dict[str, xr.DataArray]:
    """
    Concat into a single dask array all the dask arrays from each time
    chunk to make the final arrays of the xds.

    Parameters
    ----------
    dims : List[str]
        dimension names
    tvars : Dict[str, xr.DataArray]
        variables as lists of dask arrays per time chunk
    pol_cnt : int
        len of pol axis/dim
    chan_cnt : int
        len of freq axis/dim (chan indices)

    Returns
    -------
    Dict[str, xr.DataArray]
        variables as concated dask arrays
    """

    mvars = {}
    for tvr in tvars.keys():
        data_var = tvr
        if tvr == "UVW":
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0),
                dims=dims[:2] + ["uvw_coords"],
            )
        elif len(tvars[tvr][0].shape) == 3 and (tvars[tvr][0].shape[-1] == pol_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0), dims=dims[:2] + ["pol"]
            )
        elif len(tvars[tvr][0].shape) == 3 and (tvars[tvr][0].shape[-1] == chan_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0), dims=dims[:2] + ["freq"]
            )
        else:
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0),
                dims=dims[: len(tvars[tvr][0].shape)],
            )

    return mvars


def read_flat_main_table(
    infile: str,
    ddi: Union[int, None] = None,
    scan_state: Union[Tuple[int, int], None] = None,
    rowidxs=None,
    ignore_msv2_cols: Union[List, None] = None,
    chunks: Tuple[int, ...] = (22000, 512, 2),
) -> Tuple[xr.Dataset, Dict[str, Any], Dict[str, Any]]:
    """
    Read main table using flat structure: no baseline dimension. Vis
    dimensions are: row, freq, pol
    (experimental, perhaps to be deprecated/removed). Works but some
    features may be missing and/or flaky.

    Chunk tuple: (row, freq, pol)

    Parameters
    ----------
    infile : str

    ddi : Union[int, None] (Default value = None)

    scan_state : Union[Tuple[int, int], None] (Default value = None)

    rowidxs : np.ndarray (Default value = None)

    ignore_msv2_cols : Union[List, None] (Default value = None)

    chunks : Tuple[int, ...] (Default value = (22000, 512, 2))

    Returns
    -------
    Tuple[xr.Dataset, Dict[str, Any], Dict[str, Any]]
    """
    taql_where = f"where DATA_DESC_ID = {ddi}"
    if scan_state:
        # TODO: support additional intent/scan/subscan conditions if
        # we keep this read_flat functionality
        scans, states = scan_state
        # get row indices relative to full main table
        if type(states) == np.ndarray:
            state_ids_or = " OR STATE_ID = ".join(np.char.mod("%d", states))
            taql_where += f" AND (STATE_ID = {state_ids_or})"
        elif states is not None:
            taql_where += f" AND STATE_ID = {states}"
        elif scans is not None:
            taql_where += f" AND SCAN_NUMBER = {scans}"

    mtable = tables.table(
        infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )

    # get row indices relative to full main table
    if rowidxs is None:
        taql_rowid = f"select rowid() as ROWS from $mtable {taql_where}"
        with open_query(mtable, taql_rowid) as query_rows:
            rowidxs = query_rows.getcol("ROWS")

    nrows = len(rowidxs)
    if nrows == 0:
        mtable.close()
        return xr.Dataset(), {}, {}

    part_ids = get_partition_ids(mtable, taql_where)

    taql_cols = f"select * from $mtable {taql_where}"
    with open_query(mtable, taql_cols) as query_cols:
        cols = query_cols.colnames()
        ignore = [
            col
            for col in cols
            if (not query_cols.iscelldefined(col, 0))
            or (query_cols.coldatatype(col) == "record")
        ]
        cdata = dict(
            [
                (col, query_cols.getcol(col, 0, 1))
                for col in cols
                if (col not in ignore)
                and not (ignore_msv2_cols and col in ignore_msv2_cols)
            ]
        )
        chan_cnt, pol_cnt = [
            (cdata[cc].shape[1], cdata[cc].shape[2])
            for cc in cdata
            if len(cdata[cc].shape) == 3
        ][0]

    mtable.close()

    mvars, mcoords, bvars, xds = {}, {}, {}, xr.Dataset()
    # loop over row chunks
    for rc in range(0, nrows, chunks[0]):
        crlen = min(chunks[0], nrows - rc)  # chunk row length
        rcidxs = rowidxs[rc : rc + chunks[0]]

        # loop over each column and create delayed dask arrays
        for col in cdata.keys():
            if col not in bvars:
                bvars[col] = []

            if len(cdata[col].shape) == 1:
                delayed_array = dask.delayed(read_flat_col_chunk)(
                    infile, col, (crlen,), rcidxs, None, None
                )
                bvars[col] += [
                    dask.array.from_delayed(delayed_array, (crlen,), cdata[col].dtype)
                ]

            elif col == "UVW":
                delayed_array = dask.delayed(read_flat_col_chunk)(
                    infile, col, (crlen, 3), rcidxs, None, None
                )
                bvars[col] += [
                    dask.array.from_delayed(delayed_array, (crlen, 3), cdata[col].dtype)
                ]

            elif len(cdata[col].shape) == 2:
                pol_list = []
                dd = 1 if cdata[col].shape[1] == chan_cnt else 2
                for pc in range(0, cdata[col].shape[1], chunks[dd]):
                    plen = min(chunks[dd], cdata[col].shape[1] - pc)
                    delayed_array = dask.delayed(read_flat_col_chunk)(
                        infile, col, (crlen, plen), rcidxs, None, pc
                    )
                    pol_list += [
                        dask.array.from_delayed(
                            delayed_array, (crlen, plen), cdata[col].dtype
                        )
                    ]
                bvars[col] += [dask.array.concatenate(pol_list, axis=1)]

            elif len(cdata[col].shape) == 3:
                chan_list = []
                for cc in range(0, chan_cnt, chunks[1]):
                    clen = min(chunks[1], chan_cnt - cc)
                    pol_list = []
                    for pc in range(0, cdata[col].shape[2], chunks[2]):
                        plen = min(chunks[2], cdata[col].shape[2] - pc)
                        delayed_array = dask.delayed(read_flat_col_chunk)(
                            infile, col, (crlen, clen, plen), rcidxs, cc, pc
                        )
                        pol_list += [
                            dask.array.from_delayed(
                                delayed_array, (crlen, clen, plen), cdata[col].dtype
                            )
                        ]
                    chan_list += [dask.array.concatenate(pol_list, axis=2)]
                bvars[col] += [dask.array.concatenate(chan_list, axis=1)]

    # now concat all the dask chunks from each time to make the xds
    mvars = {}
    for kk in bvars.keys():
        data_var = kk
        if len(bvars[kk]) == 0:
            ignore += [kk]
            continue
        if kk == "UVW":
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0), dims=["row", "uvw_coords"]
            )
        elif len(bvars[kk][0].shape) == 2 and (bvars[kk][0].shape[-1] == pol_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0), dims=["row", "pol"]
            )
        elif len(bvars[kk][0].shape) == 2 and (bvars[kk][0].shape[-1] == chan_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0), dims=["row", "chan"]
            )
        else:
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0),
                dims=["row", "freq", "pol"][: len(bvars[kk][0].shape)],
            )

    mvars["time"] = xr.DataArray(
        convert_casacore_time(mvars["TIME"].values), dims=["row"]
    ).chunk({"row": chunks[0]})

    # add xds global attributes
    cc_attrs = extract_table_attributes(infile)
    attrs = {"other": {"msv2": {"ctds_attrs": cc_attrs, "bad_cols": ignore}}}
    # add per data var attributes
    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    mvars = rename_vars(mvars)
    mvars = redim_id_data_vars(mvars)
    xds = xr.Dataset(mvars, coords=mcoords)

    return xds, part_ids, attrs



def add_encoding(xds, compressor, chunks=None):
    if chunks is None:
        chunks = xds.sizes

    chunks = {**dict(xds.sizes), **chunks}  # Add missing sizes if presents.

    encoding = {}
    for da_name in list(xds.data_vars):
        if chunks:
            da_chunks = [chunks[dim_name] for dim_name in xds[da_name].sizes]
            xds[da_name].encoding = {"compressor": compressor, "chunks": da_chunks}
            # print(xds[da_name].encoding)
        else:
            xds[da_name].encoding = {"compressor": compressor}



"""Contains optimised functions to be used within other modules."""

import numpy as np
import xarray as xr
import pandas as pd


def cast_to_str(x):
    if isinstance(x, list):
        return x[0]
    else:
        return x


def get_pad_value(col_dtype: np.dtype) -> object:
    """
    Produce a padding/missing/nan value appropriate for a casacore data column
    (for when we need to pad data vars coming from columns with rows of
    variable size array values)

    Parameters
    ----------
    col_dtype : dtype
        dtype of data being loaded from a table column

    Returns
    -------
    object
        pad value ("missing" / "fill") for the type given
    """
    # Fill values for missing/NaN data in integer variables, based on usual
    # numpy fill values. See https://github.com/numpy/numpy/issues/21166,
    # https://github.com/casangi/xradio/issues/219, https://github.com/casangi/xradio/pull/177
    fill_value_int32 = np.int32(-2147483648)
    fill_value_int64 = np.int64(-9223372036854775808)

    if col_dtype == np.int32:
        return fill_value_int32
    elif col_dtype == np.int64 or col_dtype == "int":
        return fill_value_int64
    elif np.issubdtype(col_dtype, np.floating):
        return np.nan
    elif np.issubdtype(col_dtype, np.complexfloating):
        return complex(np.nan, np.nan)
    elif np.issubdtype(col_dtype, np.bool_):
        return False
    elif np.issubdtype(col_dtype, str):
        return ""
    else:
        raise RuntimeError(
            "Padding / missing value not defined for the type requested: "
            f"{col_dtype} (of type: {type(col_dtype)})"
        )


def to_list(x):
    if isinstance(x, np.ndarray):
        if x.ndim == 0:
            return [x.item()]
        return list(x)  # needed for json serialization
    elif isinstance(x, list):
        return x
    return [x]


def to_np_array(x):
    if isinstance(x, (list, np.ndarray)):
        if x.ndim == 0:
            return np.array([x.item()])
        return np.array(x)  # needed for json serialization
    return np.array([x])


def check_if_consistent(array: np.ndarray, array_name: str) -> np.ndarray:
    """_summary_

    Parameters
    ----------
    array : _type_
        _description_
    array_name : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    if array.ndim == 0:
        return array.item()

    array_unique = unique_1d(array)
    assert len(array_unique) == 1, array_name + " is not consistent."
    return array_unique[0]


def unique_1d(array: np.ndarray) -> np.ndarray:
    """
    Optimised version of np.unique for 1D arrays.

    Parameters
    ----------
    array : np.ndarray
        a 1D array of values.

    Returns
    -------
    np.ndarray
        a sorted array of unique values.

    """
    if isinstance(array, xr.core.dataarray.DataArray):
        array = array.values

    if array.ndim == 0:
        return np.array([array.item()])

    return np.sort(
        pd.unique(array)
    )  # Don't remove the sort! It will cause errors that are very difficult to detect. Specifically create_field_info_and_check_ephemeris has a TaQL query that requires this.


def pairing_function(antenna_pairs: np.ndarray) -> np.ndarray:
    """
    Pairing function to convert each array pair to a single value.

    This custom pairing function will only work if the maximum value is less
    than 2**20 and less than 2,048 if using signed 32-bit integers.

    Parameters
    ----------
    antenna_pairs : np.ndarray
        a 2D array containing antenna 1 and antenna
        2 ids, which forms a baseline.

    Returns
    -------
    np.ndarray
        a 1D array of the paired values.

    """
    return antenna_pairs[:, 0] * 2**20 + antenna_pairs[:, 1]


def inverse_pairing_function(paired_array: np.ndarray) -> np.ndarray:
    """
    Inverse pairing function to convert each paired value to an antenna pair.

    This inverse pairing function is the inverse of the custom pairing function.

    Parameters
    ----------
    paired_array : np.ndarray
        a 1D array of the paired values.

    Returns
    -------
    np.ndarray
        a 2D array containing antenna 1 and antenna 2 ids.

    """
    return np.column_stack(np.divmod(paired_array, 2**20))



import toolviper.utils.logger as logger, os
from typing import Tuple

import numpy as np
import xarray as xr

from casacore import tables


def revert_time(datetimes: np.ndarray) -> np.ndarray:
    """
    Convert time back from pandas datetime ref to casacore ref
    (reverse of read.convert_casacore_time).

    Parameters
    ----------
    datetimes : np.ndarray
        times in pandas reference

    Returns
    -------
    np.ndarray
        times converted to casacore reference

    """
    return (datetimes.astype(float) / 10**9) + 3506716800.0


#####################################
# translate numpy dtypes to casacore type strings
def type_converter(npdtype: str) -> str:
    cctype = "bad"
    if (npdtype == "int64") or (npdtype == "int32"):
        cctype = "int"
    elif npdtype == "bool":
        cctype = "bool"
    elif npdtype == "float32":
        cctype = "float"
    elif (npdtype == "float64") or (npdtype == "datetime64[ns]"):
        cctype = "double"
    elif npdtype == "complex64":
        cctype = "complex"
    elif npdtype == "complex128":
        cctype = "dcomplex"
    elif str(npdtype).startswith("<U"):
        cctype = "string"

    return cctype


####################################
# create and initialize new output table
def create_table(
    outfile: str,
    xds: xr.Dataset,
    max_rows: int,
    infile=None,
    cols=None,
    generic=False,
):
    if os.path.isdir(outfile):
        os.system("rm -fr %s" % outfile)

    # create column descriptions for table description
    ctds_attrs = {}
    try:
        ctds_attrs = xds.attrs["other"]["msv2"]["ctds_attrs"]
    except KeyError as exc:
        pass

    if cols is None:
        if ctds_attrs and "column_descriptions" in ctds_attrs:
            cols = {col: col for col in ctds_attrs["column_descriptions"]}
        else:
            cols = {var: var for var in xds.data_vars}
            # Would add all xds data vars regardless of description availability
            # +
            # list(xds.data_vars) +

    tabledesc = {}
    for col, var_name in cols.items():
        if ("column_descriptions" in ctds_attrs) and (
            col in ctds_attrs["column_descriptions"]
        ):
            coldesc = ctds_attrs["column_descriptions"][col]
            # col not in ignore_msv2_cols
            if (
                not generic
                and "DATA" in col
                and "shape" not in coldesc
                and var_name in xds.data_vars
            ):
                coldesc["shape"] = tuple(np.clip(xds[var_name].shape[1:], 1, None))

            if col == "UVW" or (
                (not "shape" in coldesc or type(coldesc["shape"]) == str)
                and var_name in xds.data_vars
            ):
                coldesc["shape"] = tuple(np.clip(xds[var_name].shape[1:], 1, None))
        else:
            coldesc = {"valueType": type_converter(xds[col].dtype)}
            if generic or (
                col == "UVW" or col == "DATA"
            ):  # will be statically shaped even if not originally
                coldesc = {"shape": tuple(np.clip(xds[col].shape[1:], 1, None))}
            elif xds[col].ndim > 1:  # make variably shaped
                coldesc = {"ndim": xds[col].ndim - 1}
            coldesc["name"] = col
            coldesc["desc"] = col
        tabledesc[col] = coldesc

        # fix the fun set of edge cases from casatestdata that cause errors
        if (
            "dataManagerType" in tabledesc[col]
            and tabledesc[col]["dataManagerType"] == "TiledShapeStMan"
        ) and (tabledesc[col]["ndim"] == 1):
            tabledesc[col]["dataManagerType"] = ""

    if generic:
        tb_tool = tables.table(
            outfile,
            tabledesc=tabledesc,
            nrow=max_rows,
            readonly=False,
            lockoptions={"option": "permanentwait"},
            ack=False,
        )
    else:
        tb_tool = tables.default_ms(outfile, tabledesc)
        tb_tool.addrows(max_rows)
        # if 'DATA_DESC_ID' in cols: tb_tool.putcol('DATA_DESC_ID',
        #    np.zeros((max_rows), dtype='int32') - 1, 0, max_rows)

    # write xds attributes to table keywords, skipping certain reserved attributes
    existing_keywords = tb_tool.getkeywords()
    for attr in ctds_attrs:
        if attr in [
            "other",
            "history",
            "info",
        ] + list(existing_keywords.keys()):
            continue
        tb_tool.putkeyword(attr, ctds_attrs[attr])
    if "info" in ctds_attrs:
        tb_tool.putinfo(ctds_attrs["info"])

    # copy subtables and add to main table
    if infile:
        subtables = [
            ss.path
            for ss in os.scandir(infile)
            if ss.is_dir() and ("SORTED_TABLE" not in ss.path)
        ]
        os.system("cp -r %s %s" % (" ".join(subtables), outfile))
        for subtable in subtables:
            if not tables.tableexists(
                os.path.join(outfile, subtable[subtable.rindex("/") + 1 :])
            ):
                continue
            sub_tbl = tables.table(
                os.path.join(outfile, subtable[subtable.rindex("/") + 1 :]),
                readonly=False,
                lockoptions={"option": "permanentwait"},
                ack=False,
            )
            tb_tool.putkeyword(
                subtable[subtable.rindex("/") + 1 :], sub_tbl, makesubrecord=True
            )
            sub_tbl.close()

    tb_tool.close()


############################################################################
##
## write_generic_table() - write any xds to generic casacore table format
##
############################################################################
def write_generic_table(xds: xr.Dataset, outfile: str, subtable="", cols=None):
    """
    Write generic xds contents back to casacore table format on disk

    Parameters
    ----------
    xds : xr.Dataset

    outfile : str

    subtable : str (Default value = "")

    cols : List[str] (Default value = None)

    Returns
    -------

    """
    outfile = os.path.expanduser(outfile)
    logger.debug("writing {os.path.join(outfile, subtable)}")

    try:
        ctds_attrs = {}
        ctds_attrs = xds.attrs["other"]["msv2"]["ctds_attrs"]
    except KeyError as exc:
        pass

    if cols is None:
        cols = {var.upper(): var for var in xds.data_vars}
        cols.update({coo.upper(): coo for coo in xds.coords if coo not in xds.dims})
        # Would add cols with a description regardless of presence in xds
        # + (
        #     list(
        #         ctds_attrs["column_descriptions"].keys()
        #         if "column_descriptions" in ctds_attrs
        #         else []
        #     )
        # )
    max_rows = xds.row.shape[0] if "row" in xds.dims else 0
    create_table(
        os.path.join(outfile, subtable),
        xds,
        max_rows,
        infile=None,
        cols=cols,
        generic=True,
    )

    tb_tool = tables.table(
        os.path.join(outfile, subtable),
        readonly=False,
        lockoptions={"option": "permanentwait"},
        ack=False,
    )
    try:
        for dv, col in cols.items():
            if (dv not in xds) or (np.prod(xds[dv].shape) == 0):
                continue
            values = (
                xds[dv].values
                if xds[dv].dtype != "datetime64[ns]"
                else revert_time(xds[dv].values)
            )
            tb_tool.putcol(col, values, 0, values.shape[0], 1)
    except Exception:
        print(
            "ERROR: exception in write generic table - %s, %s, %s, %s"
            % (os.path.join(outfile, subtable), dv, str(values.shape), tb_tool.nrows())
        )

    # now we have to add this subtable to the main table keywords (assuming a main table already exists)
    if len(subtable) > 0:
        main_tbl = tables.table(
            outfile, readonly=False, lockoptions={"option": "permanentwait"}, ack=False
        )
        main_tbl.putkeyword(subtable, tb_tool, makesubrecord=True)
        main_tbl.done()
    tb_tool.close()


###################################
# local helper
def write_main_table_slice(
    xda: xr.DataArray,
    outfile: str,
    ddi: int,
    col: str,
    full_shape: Tuple,
    starts: Tuple,
):
    """
    Write an xds row chunk to the corresponding main table slice

    Parameters
    ----------
    xda : xr.DataArray

    outfile : str

    ddi : int

    col : str

    full_shape : Tuple

    starts : Tuple


    Returns
    -------

    """
    # trigger the DAG for this chunk and return values while the table is unlocked
    values = xda.compute().values
    if xda.dtype == "datetime64[ns]":
        values = revert_time(values)

    tbs = tables.table(
        outfile, readonly=False, lockoptions={"option": "permanentwait"}, ack=True
    )

    # try:
    if (
        (values.ndim == 1) or (col == "UVW") or (values.shape[1:] == full_shape)
    ):  # scalar columns
        tbs.putcol(col, values, starts[0], len(values))
    else:
        if not tbs.iscelldefined(col, starts[0]):
            tbs.putcell(col, starts[0] + np.arange(len(values)), np.zeros((full_shape)))
        tbs.putcolslice(
            col,
            values,
            starts[1 : values.ndim],
            tuple(np.array(starts[1 : values.ndim]) + np.array(values.shape[1:]) - 1),
            [],
            starts[0],
            len(values),
            1,
        )
    # except:
    #    print("ERROR: write exception - %s, %s, %s" % (col, str(values.shape), str(starts)))
    tbs.close()



import copy
import os

import dask.array as da
import numpy as np
import xarray as xr
from astropy.coordinates import Angle
from casacore import tables

from .common import _active_mask, _create_new_image, _object_name, _pointing_center
from ..common import _aperture_or_sky, _compute_sky_reference_pixel, _doppler_types
from ...._utils._casacore.tables import open_table_rw


def _compute_direction_dict(xds: xr.Dataset) -> dict:
    """
    Given xds metadata, compute the direction dict that is valid
    for a CASA image coordinate system
    """
    direction = {}
    xds_dir = xds.attrs["direction"]
    direction["system"] = xds_dir["reference"]["equinox"]
    direction["projection"] = xds_dir["projection"]
    direction["projection_parameters"] = xds_dir["projection_parameters"]
    direction["units"] = np.array(xds_dir["reference"]["units"], dtype="<U16")
    direction["crval"] = np.array(xds_dir["reference"]["value"])
    direction["cdelt"] = np.array((xds.l.cdelt, xds.m.cdelt))
    direction["crpix"] = _compute_sky_reference_pixel(xds)
    direction["pc"] = np.array(xds_dir["pc"])
    direction["axes"] = ["Right Ascension", "Declination"]
    direction["conversionSystem"] = direction["system"]
    for s in ["longpole", "latpole"]:
        # longpole, latpole are numerical values in degrees in casa images
        direction[s] = Angle(str(xds_dir[s]["value"]) + xds_dir[s]["units"]).deg
    return direction


def _compute_linear_dict(xds: xr.Dataset) -> dict:
    linear = {}
    u = xds["u"].attrs
    v = xds["v"].attrs
    linear["crval"] = np.array([u["crval"], v["crval"]])
    linear["cdelt"] = np.array([u["cdelt"], v["cdelt"]])
    linear["axes"] = ["UU", "VV"]
    linear["units"] = [u["units"], v["units"]]
    lu = len(xds.coords["u"])
    lv = len(xds.coords["v"])
    linear["crpix"] = np.array([0.0, 0.0], dtype=np.float64)
    # np.interp() appears to require the x coordinate be monotonically increasing
    for i, w, z in zip([0, 1], [u, v], ["u", "v"]):
        x = xds.coords[z].values
        y = np.array(range(len(x)), dtype=np.float64)
        if w["cdelt"] < 0:
            x = x[::-1]
            y = y[::-1]
        linear["crpix"][i] = np.interp(0, x, y)
    linear["pc"] = np.array([[1.0, 0.0], [0.0, 1.0]])
    return linear


def _compute_spectral_dict(xds: xr.Dataset) -> dict:
    """
    Given xds metadata, compute the spectral dict that is valid
    for a CASA image coordinate system
    """
    spec = {}
    spec["formatUnit"] = ""
    spec["name"] = "Frequency"
    # spec["nativeType"] = _native_types.index(xds.frequency.attrs["native_type"])
    # FREQ
    spec["nativeType"] = 0
    spec["restfreq"] = xds.frequency.attrs["rest_frequency"]["value"]
    # spec["restfreqs"] = copy.deepcopy(xds.frequency.attrs["restfreqs"]["value"])
    spec["restfreqs"] = [spec["restfreq"]]
    spec["system"] = xds.frequency.attrs["frame"]
    spec["unit"] = xds.frequency.attrs["units"]
    spec["velType"] = _doppler_types.index(xds.velocity.attrs["doppler_type"])
    spec["velUnit"] = xds.velocity.attrs["units"]
    spec["version"] = 2
    spec["waveUnit"] = xds.frequency.attrs["wave_unit"]
    wcs = {}
    wcs["ctype"] = "FREQ"
    wcs["pc"] = 1.0
    wcs["crval"] = xds.frequency.attrs["crval"]
    wcs["cdelt"] = xds.frequency.attrs["cdelt"]
    wcs["crpix"] = (wcs["crval"] - xds.frequency.values[0]) / wcs["cdelt"]
    spec["wcs"] = wcs
    return spec


def _coord_dict_from_xds(xds: xr.Dataset) -> dict:
    coord = {}
    coord["telescope"] = xds.attrs["telescope"]["name"]
    coord["observer"] = xds.attrs["observer"]
    obsdate = {}
    obsdate["refer"] = xds.coords["time"].attrs["scale"]
    obsdate["type"] = "epoch"
    obsdate["m0"] = {}
    obsdate["m0"]["unit"] = xds.coords["time"].attrs["units"]
    obsdate["m0"]["value"] = xds.coords["time"].values[0]
    coord["obsdate"] = obsdate
    coord["pointingcenter"] = xds.attrs[_pointing_center].copy()
    if "position" in xds.attrs["telescope"]:
        telpos = {}
        telpos["refer"] = xds.attrs["telescope"]["position"]["ellipsoid"]
        if xds.attrs["telescope"]["position"]["ellipsoid"] == "GRS80":
            telpos["refer"] = "ITRF"
        for i in range(3):
            telpos[f"m{i}"] = {
                "unit": xds.attrs["telescope"]["position"]["units"][i],
                "value": xds.attrs["telescope"]["position"]["value"][i],
            }
        telpos["type"] = "position"
        coord["telescopeposition"] = telpos
    if "l" in xds.coords:
        coord["direction0"] = _compute_direction_dict(xds)
    else:
        coord["linear0"] = _compute_linear_dict(xds)
    coord["stokes1"] = {
        "axes": np.array(["Stokes"], dtype="<U16"),
        "cdelt": np.array([1.0]),
        "crpix": np.array([0.0]),
        "crval": np.array([1.0]),
        "pc": np.array([[1.0]]),
        "stokes": np.array(xds.polarization.values, dtype="<U16"),
    }
    coord["spectral2"] = _compute_spectral_dict(xds)
    coord["pixelmap0"] = np.array([0, 1])
    coord["pixelmap1"] = np.array([2])
    coord["pixelmap2"] = np.array([3])
    coord["pixelreplace0"] = np.array([0.0, 0.0])
    coord["pixelreplace1"] = np.array([0.0])
    coord["pixelreplace2"] = np.array([0.0])
    coord["worldmap0"] = np.array([0, 1], dtype=np.int32)
    coord["worldmap1"] = np.array([2], dtype=np.int32)
    coord["worldmap2"] = np.array([3], dtype=np.int32)
    # this probbably needs some verification
    coord["worldreplace0"] = [0.0, 0.0]
    coord["worldreplace1"] = np.array(coord["stokes1"]["crval"])
    coord["worldreplace2"] = np.array([xds.frequency.attrs["crval"]])
    return coord


def _history_from_xds(xds: xr.Dataset, image: str) -> None:
    nrows = len(xds.history.row) if "row" in xds.data_vars else 0
    if nrows > 0:
        # TODO need to implement nrows == 0 case
        with open_table_rw(os.sep.join([image, "logtable"])) as tb:
            tb.addrows(nrows + 1)
            for c in ["TIME", "PRIORITY", "MESSAGE", "LOCATION", "OBJECT_ID"]:
                vals = xds.history[c].values
                if c == "TIME":
                    k = time.time() + 40587 * 86400
                elif c == "PRIORITY":
                    k = "INFO"
                elif c == "MESSAGE":
                    k = (
                        "Wrote xds to "
                        + os.path.basename(image)
                        + " using cngi_io.xds_to_casa_image_2()"
                    )
                elif c == "LOCATION":
                    k = "cngi_io.xds_to_casa_image_2"
                elif c == "OBJECT_ID":
                    k = ""
                vals = np.append(vals, k)
                tb.putcol(c, vals)


def _imageinfo_dict_from_xds(xds: xr.Dataset) -> dict:
    ii = {}
    ap_sky = _aperture_or_sky(xds)
    ii["image_type"] = (
        xds[ap_sky].attrs["image_type"] if "image_type" in xds[ap_sky].attrs else ""
    )
    ii["objectname"] = xds.attrs[_object_name]
    if "beam" in xds.data_vars:
        # multi beam
        pp = {}
        pp["nChannels"] = len(xds.frequency)
        pp["nStokes"] = len(xds.polarization)
        bu = xds.beam.attrs["units"]
        chan = 0
        polarization = 0
        bv = xds.beam.values
        for i in range(pp["nChannels"] * pp["nStokes"]):
            bp = bv[0][polarization][chan][:]
            b = {
                "major": {"unit": bu, "value": bp[0]},
                "minor": {"unit": bu, "value": bp[1]},
                "positionangle": {"unit": bu, "value": bp[2]},
            }
            pp["*" + str(pp["nChannels"] * polarization + chan)] = b
            chan += 1
            if chan >= pp["nChannels"]:
                chan = 0
                polarization += 1
        ii["perplanebeams"] = pp
    elif "beam" in xds.attrs and xds.attrs["beam"]:
        # do nothing if xds.attrs['beam'] is None
        ii["restoringbeam"] = copy.deepcopy(xds.attrs["beam"])
        for k in ["major", "minor", "pa"]:
            del ii["restoringbeam"][k]["type"]
            ii["restoringbeam"][k]["unit"] = ii["restoringbeam"][k]["units"]
            del ii["restoringbeam"][k]["units"]
        ii["restoringbeam"]["positionangle"] = copy.deepcopy(ii["restoringbeam"]["pa"])
        del ii["restoringbeam"]["pa"]
    return ii


def _write_casa_data(xds: xr.Dataset, image_full_path: str) -> None:
    sky_ap = _aperture_or_sky(xds)
    if xds[sky_ap].shape[0] != 1:
        raise RuntimeError("XDS can only be converted if it has exactly one time plane")
    trans_coords = (
        ("frequency", "polarization", "m", "l")
        if sky_ap == "SKY"
        else ("frequency", "polarization", "v", "u")
    )
    casa_image_shape = xds[sky_ap].isel(time=0).transpose(*trans_coords).shape[::-1]
    active_mask = xds.attrs["active_mask"] if _active_mask in xds.attrs else ""
    masks = []
    masks_rec = {}
    mask_rec = {
        "box": {
            "blc": np.array([1.0, 1.0, 1.0, 1.0]),
            "comment": "",
            "isRegion": 1,
            "name": "LCBox",
            "oneRel": True,
            "shape": np.array(casa_image_shape),
            "trc": np.array(casa_image_shape),
        },
        "comment": "",
        "isRegion": 1,
        "name": "LCPagedMask",
    }
    for m in xds.data_vars:
        attrs = xds[m].attrs
        if "image_type" in attrs and attrs["image_type"] == "Mask":
            masks_rec[m] = mask_rec
            masks_rec[m]["mask"] = f"Table: {os.sep.join([image_full_path, m])}"
            masks.append(m)
    myvars = [sky_ap]
    myvars.extend(masks)
    # xr.apply_ufunc seems to like stripping attributes from xds and its coordinates,
    # so make a deep copy of the pertinent thing to be handled in xr.apply_ufunc()
    nan_mask = xr.apply_ufunc(da.isnan, xds[sky_ap].copy(deep=True), dask="allowed")
    # test if nan pixels are all already masked in active_mask.
    # if so, don't worry about masking them again, just use active_mask
    # which already masks all nans
    arr_masks = {}
    do_mask_nans = False
    there_are_nans = nan_mask.any()
    if there_are_nans:
        has_active_mask = bool(active_mask)
        if not has_active_mask:
            do_mask_nans = True
        else:
            notted_active_mask = xr.apply_ufunc(
                da.logical_not, xds[active_mask].copy(deep=True), dask="allowed"
            )
            some_nans_are_not_already_masked = xr.apply_ufunc(
                da.logical_and, nan_mask, notted_active_mask, dask="allowed"
            )
            do_mask_nans = some_nans_are_not_already_masked.any()
    if do_mask_nans:
        mask_name = "mask_xds_nans"
        i = 0
        while mask_name in masks:
            mask_name = f"mask_xds_nans{i}"
            i += 1
        masks_rec[mask_name] = mask_rec
        masks_rec[mask_name][
            "mask"
        ] = f"Table: {os.sep.join([image_full_path, mask_name])}"
        masks.append(mask_name)
        arr_masks[mask_name] = nan_mask
        if active_mask:
            mask_name = f"mask_xds_nans_or_{active_mask}"
            i = 0
            while mask_name in masks:
                mask_name = f"mask_xds_nans{i}"
                i += 1
            masks_rec[mask_name] = mask_rec
            masks_rec[mask_name][
                "mask"
            ] = f"Table: {os.sep.join([image_full_path, mask_name])}"
            masks.append(mask_name)
            # This command strips attributes at various places for no
            # apparent reason, so make a copy of the xds to run it on
            arr_masks[mask_name] = xr.apply_ufunc(
                da.logical_or,
                xds[active_mask].copy(deep=True),
                nan_mask,
                dask="allowed",
            )
        active_mask = mask_name
    data_type = "complex" if "u" in xds.coords else "float"
    _write_initial_image(xds, image_full_path, active_mask, casa_image_shape[::-1])
    for v in myvars:
        _write_pixels(v, active_mask, image_full_path, xds)
    for name, v in arr_masks.items():
        _write_pixels(name, active_mask, image_full_path, xds, v)
    if masks:
        with open_table_rw(image_full_path) as tb:
            tb.putkeyword("masks", masks_rec)
            tb.putkeyword("Image_defaultmask", active_mask)


def _write_initial_image(
    xds: xr.Dataset, imagename: str, maskname: str, image_shape: tuple
) -> None:
    if not maskname:
        maskname = ""
    for dv in ["SKY", "APERTURE"]:
        if dv in xds.data_vars:
            value = xds[dv][0, 0, 0, 0, 0].values.item()
            if xds[dv][0, 0, 0, 0, 0].values.dtype == "float32":
                value = "default"
            break
    # print(type(value))
    image_full_path = os.path.expanduser(imagename)
    with _create_new_image(
        image_full_path, mask=maskname, shape=image_shape, value=value
    ) as casa_image:
        # just create the image, don't do anythong with it
        pass


def _write_image_block(xda: xr.DataArray, outfile: str, blc: tuple) -> None:
    """
    Write image xda chunk to the corresponding image table slice
    """
    # trigger the DAG for this chunk and return values while the table is
    # unlocked
    values = xda.compute().values
    with open_table_rw(outfile) as tb_tool:
        tb_tool.putcellslice(
            tb_tool.colnames()[0],
            0,
            values,
            blc,
            tuple(np.array(blc) + np.array(values.shape) - 1),
        )


def _write_pixels(
    v: str,
    active_mask: str,
    image_full_path: str,
    xds: xr.Dataset,
    value: xr.DataArray = None,
) -> None:
    flip = False
    if v == "SKY" or v == "APERTURE":
        filename = image_full_path
    else:
        # mask
        flip = True
        filename = os.sep.join([image_full_path, v])
        # the default mask has already been written in _xds_to_casa_image()
        if not os.path.exists(filename):
            tb = tables.table(os.sep.join([image_full_path, active_mask]))
            tb.copy(filename, deep=True, valuecopy=True)
            tb.close()
    if "l" in xds.coords:
        trans_coords = ("frequency", "polarization", "m", "l")
    elif "u" in xds.coords:
        trans_coords = ("frequency", "polarization", "v", "u")
    else:
        raise RuntimeError(f"Unhandled coords combination {xds.coords.keys()}")
    trans_coord = ("frequency", "polarization", "m", "l")
    arr = xds[v] if v in xds.data_vars else value
    arr = arr.isel(time=0).transpose(*trans_coords)
    chunk_bounds = tuple(zip(arr.shape)) if arr.chunks is None else arr.chunks
    b = [0, 0, 0, 0]
    loc0, loc1, loc2, loc3 = (0, 0, 0, 0)
    for i0 in chunk_bounds[0]:
        b[0] = loc0
        s0 = slice(b[0], b[0] + i0)
        loc1 = 0
        for i1 in chunk_bounds[1]:
            b[1] = loc1
            s1 = slice(b[1], b[1] + i1)
            loc2 = 0
            for i2 in chunk_bounds[2]:
                b[2] = loc2
                s2 = slice(b[2], b[2] + i2)
                loc3 = 0
                for i3 in chunk_bounds[3]:
                    b[3] = loc3
                    blc = tuple(b)
                    s3 = slice(b[3], b[3] + i3)
                    sub_arr = arr[s0, s1, s2, s3]
                    if flip:
                        sub_arr = np.logical_not(sub_arr)
                    _write_image_block(sub_arr, filename, blc)
                    loc3 += i3
                loc2 += i2
            loc1 += i1
        loc0 += i0



#################################
#
# Public interface
#
#################################
import warnings
from typing import List, Union
import copy
import numpy as np
import xarray as xr

# from .._utils.zarr.common import _load_no_dask_zarr

from ._util.casacore import _load_casa_image_block, _xds_to_casa_image
from ._util.fits import _read_fits_image
from ._util.image_factory import (
    _make_empty_aperture_image,
    _make_empty_lmuv_image,
    _make_empty_sky_image,
)
from ._util.zarr import _load_image_from_zarr_no_dask, _xds_from_zarr, _xds_to_zarr

warnings.filterwarnings("ignore", category=FutureWarning)


def read_image(
    infile: str,
    chunks: dict = {},
    verbose: bool = False,
    do_sky_coords: bool = True,
    selection: dict = {},
) -> xr.Dataset:
    """
    Convert CASA, FITS, or zarr image to xradio image xds format
    ngCASA image spec is located at
    https://docs.google.com/spreadsheets/d/1WW0Gl6z85cJVPgtdgW4dxucurHFa06OKGjgoK8OREFA/edit#gid=1719181934

    Parameters
    ----------
    infile : str
        Path to the input CASA image
    :chunks : dict
        The desired dask chunk size. Only applicable for casacore and fits images.
        Supported optional keys are 'l', 'm', 'frequency', 'polarization', and 'time'.
        The supported values are positive integers, indicating the length of a chunk
        on that particular axis. If a key is missing, then the associated chunk length
        along that axis is equal to the number of pixels along that axis. For zarr
        images, this parameter is ignored and the chunk size used to store the arrays
        in the zarr image is used. 'l' represents the longitude like dimension, and 'm'
        represents the latitude like dimension. For apeature images, 'u' may be used in
        place of 'l', and 'v' in place of 'm'.
    verbose : bool
        emit debugging messages? Default is False.
    do_sky_coords : bool
        Compute SkyCoord at each pixel and add spherical (sky) dimensions as non-dimensional
        coordinates in the returned xr.Dataset. Only applies to CASA and FITS images; zarr
        images will have these coordinates added if they were saved with the zarr dataset,
        and if zarr image didn't have these coordinates when it was written, the resulting
        xr.Dataset will not.
    selection : dict
        The selection of data to return, supported keys are time,
        polarization, frequency, l (or u if aperture image), m (or v if aperture
        image) a missing key indicates to return the entire axis length for that
        dimension. Values can be non-negative integers or slices. Slicing
        behaves as numpy slicing does, that is the start pixel is included in
        the selection, and the end pixel is not. An empty dictionary (the
        default) indicates that the entire image should be returned. Currently
        only supported for images stored in zarr format.

    Returns
    -------
    xarray.Dataset
    """
    emsgs = []
    do_casa = True
    try:
        from ._util.casacore import _read_casa_image
    except Exception as e:
        emsgs.append(
            "python-casacore could not be imported, will not try to "
            f"read as casacore image: {e.args}"
        )
        do_casa = False
    if do_casa:
        # next statement is for debug
        # return _read_casa_image(infile, chunks, verbose, do_sky_coords)
        try:
            return _read_casa_image(infile, chunks, verbose, do_sky_coords)
        except Exception as e:
            emsgs.append(f"image format appears not to be casacore: {e.args}")
    # next statement is for debug, comment when done debugging
    # return _read_fits_image(infile, chunks, verbose, do_sky_coords)
    try:
        return _read_fits_image(infile, chunks, verbose, do_sky_coords)
    except Exception as e:
        emsgs.append(f"image format appears not to be fits {e.args}")
    # when done debuggin comment out next line
    # return _xds_from_zarr(infile, {"dv": "dask", "coords": "numpy"}, selection=selection)
    try:
        return _xds_from_zarr(
            infile, {"dv": "dask", "coords": "numpy"}, selection=selection
        )
    except Exception as e:
        emsgs.append(f"image format appears not to be zarr {e.args}")
    emsgs.insert(
        0, f"Unrecognized image format. Supported types are CASA, FITS, and zarr.\n"
    )
    raise RuntimeError("\n".join(emsgs))


def load_image(infile: str, block_des: dict = {}, do_sky_coords=True) -> xr.Dataset:
    """
    Load an image or portion of an image (subimage) into memory with data variables
    being converted from dask to numpy arrays and coordinate arrays being converted
    from dask arrays to numpy arrays. If already a numpy array, that data variable
    or coordinate is left unaltered.

    Parameters
    ----------
    infile : str
        Path to the input image, currently CASA and zarr images are supported
    block_des : dict
        The description of data to return, supported keys are time,
        polarization, frequency, l (or u if aperture image), m (or v if aperture
        image) a missing key indicates to return the entire axis length for that
        dimension. Values can be non-negative integers or slices. Slicing
        behaves as numpy slicing does, that is the start pixel is included in
        the selection, and the end pixel is not. An empty dictionary (the
        default) indicates that the entire image should be returned. The returned
        dataset will have data variables stored as numpy, not dask, arrays.
        TODO I'd really like to rename this parameter "selection"
    do_sky_coords : bool
        Compute SkyCoord at each pixel and add spherical (sky) dimensions as non-dimensional
        coordinates in the returned xr.Dataset. Only applies to CASA and FITS images; zarr
        images will have these coordinates added if they were saved with the zarr dataset,
        and if zarr image didn't have these coordinates when it was written, the resulting
        xr.Dataset will not.
    Returns
    -------
    xarray.Dataset
    """
    do_casa = True
    emsgs = []
    selection = copy.deepcopy(block_des) if block_des else block_des
    if selection:
        for k, v in selection.items():
            if type(v) == int:
                selection[k] = slice(v, v + 1)
    try:
        from ._util.casacore import _read_casa_image
    except Exception as e:
        emsgs.append(
            "python-casacore could not be imported, will not try to "
            f"read as casacore image: {e.args}"
        )
        do_casa = False
    if do_casa:
        # comment next line when done debugging
        # return _load_casa_image_block(infile, bd, do_sky_coords)
        try:
            return _load_casa_image_block(infile, selection, do_sky_coords)
        except Exception as e:
            emsgs.append(f"image format appears not to be casacore: {e.args}")
    """
    try:
        return __read_fits_image(infile, chunks, masks, history, verbose)
    except Exception as e:
        emsgs.append(f'image format appears not to be fits {e.args}')
    """
    # when done debugging, comment out next line
    # return _load_image_from_zarr_no_dask(infile, block_des)
    # return _xds_from_zarr(infile, {"dv": "numpy"}, selection)
    try:
        return _load_image_from_zarr_no_dask(infile, block_des)
        # return _xds_from_zarr(infile, {"dv": "numpy", "coords": "numpy"}, selection)
    except Exception as e:
        emsgs.append(f"image format appears not to be zarr {e.args}")
    emsgs.insert(
        0, f"Unrecognized image format. Supported formats are casacore and zarr.\n"
    )
    raise RuntimeError("\n".join(emsgs))


def write_image(
    xds: xr.Dataset, imagename: str, out_format: str = "casa", overwrite=False
) -> None:
    """
    Convert an xds image to CASA or zarr image.
    xds : xarray.Dataset
        XDS to convert
    imagename : str
        Path to output CASA image
    out_format : str
        Format of output image, currently "casa" and "zarr" are supported
    Returns
    -------
    None
    """
    my_format = out_format.lower()

    if overwrite:
        import os

        os.system("rm -rf " + imagename)

    if my_format == "casa":
        _xds_to_casa_image(xds, imagename)
    elif my_format == "zarr":
        _xds_to_zarr(xds, imagename)
    else:
        raise ValueError(
            f"Writing to format {out_format} is not supported. "
            'out_format must be either "casa" or "zarr".'
        )


def make_empty_sky_image(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str = "FK5",
    projection: str = "SIN",
    spectral_reference: str = "lsrk",
    do_sky_coords: bool = True,
) -> xr.Dataset:
    """
    Create an image xarray.Dataset with only coordinates (no datavariables).
    The image dimensionality is time, pol, chan, l, m

    Parameters
    ----------
    phase_center : array of float, length = 2, units = rad
        Image phase center.
    image_size : array of int, length = 2
        Number of x and y axis pixels in image.
    cell_size : array of float, length = 2, units = rad
        Cell size of x and y axis pixels in image.
    chan_coords : list or np.ndarray
        The center frequency in Hz of each image channel.
    pol_coords : list or np.ndarray
        The polarization code for each image polarization.
    time_coords : list or np.ndarray
        The time for each temporal plane in MJD.
    direction_reference : str, default = 'FK5'
    projection : str, default = 'SIN'
    spectral_reference : str, default = 'lsrk'
    do_sky_coords : bool
        If True, compute SkyCoord at each pixel and add spherical (sky) dimensions as
        non-dimensional coordinates in the returned xr.Dataset.
    Returns
    -------
    xarray.Dataset
    """
    return _make_empty_sky_image(
        phase_center,
        image_size,
        cell_size,
        chan_coords,
        pol_coords,
        time_coords,
        direction_reference,
        projection,
        spectral_reference,
        do_sky_coords,
    )


def make_empty_aperture_image(
    phase_center: Union[List[float], np.ndarray],
    image_size: Union[List[int], np.ndarray],
    sky_image_cell_size: Union[List[float], np.ndarray],
    chan_coords: Union[List[float], np.ndarray],
    pol_coords: Union[List[str], np.ndarray],
    time_coords: Union[List[float], np.ndarray],
    direction_reference: str = "FK5",
    projection: str = "SIN",
    spectral_reference: str = "lsrk",
) -> xr.Dataset:
    """
    Create an aperture (uv) mage xarray.Dataset with only coordinates (no datavariables).
    The image dimensionality is time, pol, chan, u, v

    Parameters
    ----------
    phase_center : array of float, length = 2, units = rad
        Image phase center.
    image_size : array of int, length = 2
        Number of x and y axis pixels in image.
    sky_image_cell_size : array of float, length = 2, units = rad
        Cell size of x and y axis pixels in sky image, used to get cell size in uv image
    chan_coords : list or np.ndarray
        The center frequency in Hz of each image channel.
    pol_coords : list or np.ndarray
        The polarization code for each image polarization.
    time_coords : list or np.ndarray
        The time for each temporal plane in MJD.
    direction_reference : str, default = 'FK5'
    projection : str, default = 'SIN'
    spectral_reference : str, default = 'lsrk'
    Returns
    -------
    xarray.Dataset
    """
    return _make_empty_aperture_image(
        phase_center,
        image_size,
        sky_image_cell_size,
        chan_coords,
        pol_coords,
        time_coords,
        direction_reference,
        projection,
        spectral_reference,
    )


def make_empty_aperture_image(
    phase_center: Union[List[float], np.ndarray],
    image_size: Union[List[int], np.ndarray],
    sky_image_cell_size: Union[List[float], np.ndarray],
    chan_coords: Union[List[float], np.ndarray],
    pol_coords: Union[List[str], np.ndarray],
    time_coords: Union[List[float], np.ndarray],
    direction_reference: str = "FK5",
    projection: str = "SIN",
    spectral_reference: str = "lsrk",
) -> xr.Dataset:
    """
    Create an aperture (uv) mage xarray.Dataset with only coordinates (no datavariables).
    The image dimensionality is time, pol, chan, u, v

    Parameters
    ----------
    phase_center : array of float, length = 2, units = rad
        Image phase center.
    image_size : array of int, length = 2
        Number of x and y axis pixels in image.
    sky_image_cell_size : array of float, length = 2, units = rad
        Cell size of x and y axis pixels in sky image, used to get cell size in uv image
    chan_coords : list or np.ndarray
        The center frequency in Hz of each image channel.
    pol_coords : list or np.ndarray
        The polarization code for each image polarization.
    time_coords : list or np.ndarray
        The time for each temporal plane in MJD.
    direction_reference : str, default = 'FK5'
    projection : str, default = 'SIN'
    spectral_reference : str, default = 'lsrk'
    Returns
    -------
    xarray.Dataset
    """
    return _make_empty_aperture_image(
        phase_center,
        image_size,
        sky_image_cell_size,
        chan_coords,
        pol_coords,
        time_coords,
        direction_reference,
        projection,
        spectral_reference,
    )


def make_empty_lmuv_image(
    phase_center: Union[List[float], np.ndarray],
    image_size: Union[List[int], np.ndarray],
    sky_image_cell_size: Union[List[float], np.ndarray],
    chan_coords: Union[List[float], np.ndarray],
    pol_coords: Union[List[float], np.ndarray],
    time_coords: Union[List[float], np.ndarray],
    direction_reference: str = "FK5",
    projection: str = "SIN",
    spectral_reference: str = "lsrk",
    do_sky_coords: bool = True,
) -> xr.Dataset:
    """
    Create an image xarray.Dataset with only coordinates (no datavariables).
    The image dimensionality is time, pol, chan, l, m, u, v

    Parameters
    ----------
    phase_center : array of float, length = 2, units = rad
        Image phase center.
    image_size : array of int, length = 2
        Number of x and y axis pixels in image.
    sky_image_cell_size : array of float, length = 2, units = rad
        Cell size of sky image. The cell size of the u,v image will be computed from
        1/(image_size * sky_image_cell_size)
    chan_coords : list or np.ndarray
        The center frequency in Hz of each image channel.
    pol_coords : list or np.ndarray
        The polarization code for each image polarization.
    time_coords : list or np.ndarray
        The time for each temporal plane in MJD.
    direction_reference : str, default = 'FK5'
    projection : str, default = 'SIN'
    spectral_reference : str, default = 'lsrk'
    do_sky_coords : bool
        If True, compute SkyCoord at each pixel and add spherical (sky) dimensions as
        non-dimensional coordinates in the returned xr.Dataset.
    Returns
    -------
    xarray.Dataset
    """
    return _make_empty_lmuv_image(
        phase_center,
        image_size,
        sky_image_cell_size,
        chan_coords,
        pol_coords,
        time_coords,
        direction_reference,
        projection,
        spectral_reference,
        do_sky_coords,
    )



import numpy as np
import xarray as xr

from casacore import tables
import toolviper.utils.logger as logger

from .subtables import subt_rename_ids
from ._tables.read import load_generic_table, convert_casacore_time
from xradio._utils.list_and_array import check_if_consistent, unique_1d, to_list


def create_info_dicts(
    in_file: str,
    xds: xr.Dataset,
    field_and_source_xds: xr.Dataset,
    partition_info_misc_fields: dict,
    tb_tool: tables.table,
) -> dict:
    """
    For an MSv4, produces several info dicts (partition_info, processor_info,
    observation_info). The info dicts are returned in a dictionary that
    contains them indexed by their corresponding keys, which can be used
    directly to update the attrs dict of an MSv4.

    Parameters:
    -----------
    in_file: str
       path to the input MSv2
    xds: xr.Dataset
       main xds of the MSv4 being converted
    field_and_source_xds: xr.Dataset
       field_and_source_xds subdataset
    partition_info_misc_fiels: dict
       dict with several scalar fields for the partition_info dict that are
       collected while processing the main MSv4 table. Expected: scan_id,
       obs_mode, taql_where
    tb_tool: tables.table
       table (query) on the main table with an MSv4 query

    Returns:
    --------
    info_dicts: dict
       info dicts ready to be used to update the attrs of the MSv4
    """

    if "line_name" in field_and_source_xds.coords:
        line_name = to_list(unique_1d(np.ravel(field_and_source_xds.line_name.values)))
    else:
        line_name = []

    info_dicts = {}
    info_dicts["partition_info"] = {
        # "spectral_window_id": xds.frequency.attrs["spectral_window_id"],
        "spectral_window_name": xds.frequency.attrs["spectral_window_name"],
        # "field_id": to_list(unique_1d(field_id)),
        "field_name": to_list(np.unique(field_and_source_xds.field_name.values)),
        "polarization_setup": to_list(xds.polarization.values),
        "scan_number": to_list(np.unique(partition_info_misc_fields["scan_id"])),
        "source_name": to_list(np.unique(field_and_source_xds.source_name.values)),
        # "source_id": to_list(unique_1d(source_id)),
        "intents": partition_info_misc_fields["intents"].split(","),
        "taql": partition_info_misc_fields["taql_where"],
        "line_name": line_name,
    }

    observation_id = check_if_consistent(
        tb_tool.getcol("OBSERVATION_ID"), "OBSERVATION_ID"
    )
    info_dicts["observation_info"] = create_observation_info(in_file, observation_id)

    processor_id = check_if_consistent(tb_tool.getcol("PROCESSOR_ID"), "PROCESSOR_ID")
    info_dicts["processor_info"] = create_processor_info(in_file, processor_id)

    return info_dicts


def create_observation_info(in_file: str, observation_id: int):
    """
    Makes a dict with the observation info extracted from the PROCESSOR subtable.

    Parameters
    ----------
    in_file: str
       path to an input MSv2
    processor_id: int
        processor ID for one MSv4 dataset

    Returns:
    --------
    observation_info: dict
        observation description ready for the MSv4 observation_info attr
    """

    generic_observation_xds = load_generic_table(
        in_file,
        "OBSERVATION",
        rename_ids=subt_rename_ids["OBSERVATION"],
        taql_where=f" where ROWID() = {observation_id}",
    )

    observation_info = {
        "observer": [generic_observation_xds["OBSERVER"].values[0]],
        "release_date": str(
            convert_casacore_time(generic_observation_xds["RELEASE_DATE"].values)[0]
        ),
    }
    # could just assume lower:upper case but keeping explicit dict for now
    mandatory_fields = {"project": "PROJECT"}
    for field_msv4, row_msv2 in mandatory_fields.items():
        observation_info[field_msv4] = generic_observation_xds[row_msv2].values[0]

    exec_block_xds = None
    try:
        exec_block_xds = load_generic_table(in_file, "ASDM_EXECBLOCK")
    except ValueError as exc:
        logger.debug(
            "Did not find the ASDM_EXECBLOCK subtable, not loading optional fields in observation_info"
        )
    if exec_block_xds:
        exec_block_info = extract_exec_block_info(exec_block_xds)
        observation_info.update(exec_block_info)

    return observation_info


def extract_exec_block_info(exec_block_xds: xr.Dataset) -> dict:
    """
    Get the (optional) fields of the observation_info that come from the
    ASDM_EXECBLOCK subtable.

    Note this does not parse strings like 'session_reference':
    '<EntityRef entityId="uid://A001/X133d/X169f" partId="X00000000" entityTypeName="OUSStatus"'
    We might want to simplify that to 'uid://A001/X133d/X169f', but keeping the
    full string for now, as it has additional information such as the type.

    Parameters
    ----------
    exec_block_xds: xr.Dataset
        raw xds read from subtable ASDM_EXECBLOCK

    Returns:
    --------
    exec_block_info: dict
        Execution block description ready for the MSv4 observation_info dict
    """

    optional_fields = {
        "execution_block_id": "execBlockId",
        "execution_block_number": "execBlockNum",
        "execution_block_UID": "execBlockUID",
        "session_reference": "sessionReference",
        "observing_script": "observingScript",
        "observing_script_UID": "observingScriptUID",
        "observing_log": "observingLog",
    }

    exec_block_info = {}
    for field_msv4, row_msv2 in optional_fields.items():
        if row_msv2 in exec_block_xds.data_vars:
            msv2_value = exec_block_xds[row_msv2].values[0]
            if isinstance(msv2_value, np.ndarray):
                exec_block_info[field_msv4] = ",".join([log for log in msv2_value])
            else:
                exec_block_info[field_msv4] = msv2_value

    return exec_block_info


def create_processor_info(in_file: str, processor_id: int):
    """
    Makes a dict with the processor info extracted from the PROCESSOR subtable.

    Parameters
    ----------
    in_file: str
        path to an input MSv2
    processor_id: int
        processor ID for one MSv4 dataset

    Returns:
    --------
    processor_info: dict
        processor description ready for the MSv4 processor_info attr
    """

    generic_processor_xds = load_generic_table(
        in_file,
        "PROCESSOR",
        rename_ids=subt_rename_ids["PROCESSOR"],
        taql_where=f" where ROWID() = {processor_id}",
    )

    # Many telescopes (ASKAP, MeerKAT, SKA-Mid, VLBI, VLBA, ngEHT) seem to
    # produce an empty PROCESSOR subtable
    if len(generic_processor_xds.data_vars) <= 0:
        processor_info = {"type": "", "sub_type": ""}
    else:
        processor_info = {
            "type": generic_processor_xds["TYPE"].values[0],
            "sub_type": generic_processor_xds["SUB_TYPE"].values[0],
        }

    return processor_info
